#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
import copy

import pytest

import jsonschema as js

import libnmstate
from libnmstate.schema import Constants
from libnmstate.schema import DNS
from libnmstate.schema import Ethernet
from libnmstate.schema import InfiniBand
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType
from libnmstate.schema import LinuxBridge as LB
from libnmstate.schema import MacVlan
from libnmstate.schema import MacVtap
from libnmstate.schema import OVSBridge
from libnmstate.schema import OVSInterface
from libnmstate.schema import Route
from libnmstate.schema import RouteRule
from libnmstate.schema import Team
from libnmstate.schema import VRF
from libnmstate.schema import VXLAN


INTERFACES = Constants.INTERFACES
ROUTES = Constants.ROUTES
THE_BRIDGE = "br0"
VXLAN0 = "vxlan0"
TEAM0 = "team0"
SRIOV_MAC = "12:34:56:78:90:ab"

COMMON_DATA = {
    INTERFACES: [
        {
            Interface.NAME: "lo",
            Interface.DESCRIPTION: "Loopback Interface",
            Interface.TYPE: InterfaceType.UNKNOWN,
            Interface.STATE: InterfaceState.DOWN,
            "link-speed": 1000,
            Interface.MAC: "12:34:56:78:90:ab",
            Interface.MTU: 1500,
            # Read Only entries
            "if-index": 0,
            "admin-status": "up",
            "link-status": "down",
            "phys-address": "12:34:56:78:90:ab",
            "higher-layer-if": "",
            "lower-layer-if": "",
            "low-control": True,
            "statistics": {
                "in-octets": 0,
                "in-unicast-pkts": 0,
                "in-broadcast-pkts": 0,
                "in-multicast-pkts": 0,
                "in-discards": 0,
                "in-errors": 0,
                "out-octets": 0,
                "out-unicast-pkts": 0,
                "out-broadcast-pkts": 0,
                "out-multicast-pkts": 0,
                "out-discards": 0,
                "out-errors": 0,
            },
        }
    ],
    ROUTES: {
        Route.CONFIG: [
            {
                Route.TABLE_ID: 254,
                Route.METRIC: 100,
                Route.DESTINATION: "0.0.0.0/0",
                Route.NEXT_HOP_INTERFACE: "eth0",
                Route.NEXT_HOP_ADDRESS: "192.0.2.1",
            }
        ],
        Route.RUNNING: [
            {
                Route.TABLE_ID: 254,
                Route.METRIC: 100,
                Route.DESTINATION: "::/0",
                Route.NEXT_HOP_INTERFACE: "eth0",
                Route.NEXT_HOP_ADDRESS: "fe80::1",
            }
        ],
    },
    RouteRule.KEY: {
        RouteRule.CONFIG: [
            {
                RouteRule.IP_FROM: "192.0.2.0/24",
                RouteRule.IP_TO: "198.51.100.0/24",
                RouteRule.PRIORITY: 500,
                RouteRule.ROUTE_TABLE: 254,
            }
        ]
    },
    DNS.KEY: {
        DNS.RUNNING: {
            DNS.SERVER: ["2001:db8::1", "192.0.2.1"],
            DNS.SEARCH: ["example.com", "example.org"],
        },
        DNS.CONFIG: {
            DNS.SERVER: ["2001:db8::1", "192.0.2.1"],
            DNS.SEARCH: ["example.com", "example.org"],
        },
    },
}


@pytest.fixture
def default_data():
    return copy.deepcopy(COMMON_DATA)


@pytest.fixture
def portless_bridge_state():
    return {
        Interface.NAME: THE_BRIDGE,
        Interface.STATE: InterfaceState.UP,
        Interface.TYPE: LB.TYPE,
        LB.CONFIG_SUBTREE: {LB.PORT_SUBTREE: []},
    }


@pytest.fixture
def bridge_state(portless_bridge_state):
    port = {LB.Port.NAME: "eth1", LB.Port.VLAN_SUBTREE: {}}
    portless_bridge_state[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE].append(port)
    return portless_bridge_state


@pytest.fixture
def portless_ovs_bridge_state():
    return {
        Interface.NAME: THE_BRIDGE,
        Interface.STATE: InterfaceState.UP,
        Interface.TYPE: OVSBridge.TYPE,
        LB.CONFIG_SUBTREE: {OVSBridge.PORT_SUBTREE: []},
    }


@pytest.fixture
def ovs_bridge_state(portless_ovs_bridge_state):
    port = {LB.Port.NAME: "eth1", OVSBridge.Port.VLAN_SUBTREE: {}}
    ovs_bridge_state_config = portless_ovs_bridge_state[
        OVSBridge.CONFIG_SUBTREE
    ]
    ovs_bridge_state_config[OVSBridge.PORT_SUBTREE].append(port)
    return portless_ovs_bridge_state


class TestIfaceCommon:
    def test_valid_instance(self, default_data):
        libnmstate.validator.schema_validate(default_data)

    def test_invalid_instance(self, default_data):
        default_data[INTERFACES][0][Interface.STATE] = "bad-state"

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)

    def test_invalid_type(self, default_data):
        default_data[INTERFACES][0][Interface.TYPE] = "bad-type"

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)


class TestIfaceMacAddress:
    @pytest.mark.parametrize(
        "mac_address",
        [
            "00:11:22:33",
            "00:11:22:33:44:55",
            "80:00:02:08:fe:80:00:00:00:00:00:00:f4:52:14:03:00:8d:52:11",
        ],
    )
    def test_valid_mac_address(self, default_data, mac_address):
        default_data[INTERFACES][0][Interface.MAC] = mac_address
        libnmstate.validator.schema_validate(default_data)

    @pytest.mark.parametrize(
        "mac_address",
        [
            "00:11:22",
            "00:xx:xx:yy:100",
            "90:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00"
            "00:00:00:00:00:00:00:00:00:00:00",
        ],
    )
    def test_invalid_mac_address(self, default_data, mac_address):
        default_data[INTERFACES][0][Interface.MAC] = mac_address
        with pytest.raises(js.ValidationError, match=str(mac_address)):
            libnmstate.validator.schema_validate(default_data)


class TestIfaceTypeEthernet:
    def test_valid_ethernet_with_auto_neg(self, default_data):
        default_data[INTERFACES][0].update(
            {
                Interface.TYPE: InterfaceType.ETHERNET,
                Ethernet.CONFIG_SUBTREE: {Ethernet.AUTO_NEGOTIATION: True},
            }
        )
        libnmstate.validator.schema_validate(default_data)

    def test_valid_ethernet_without_auto_neg(self, default_data):
        default_data[INTERFACES][0].update(
            {
                Ethernet.CONFIG_SUBTREE: {
                    Ethernet.AUTO_NEGOTIATION: False,
                    Ethernet.SPEED: 1000,
                    Ethernet.DUPLEX: Ethernet.FULL_DUPLEX,
                },
            }
        )
        libnmstate.validator.schema_validate(default_data)

    def test_valid_without_auto_neg_and_missing_speed(self, default_data):
        """
        Defining autonegotiation as false and not specifying the link-speed is
        not a valid configuration, however, this is not handled by the schema
        at the moment, deferring the handling to the application code.
        """
        default_data[INTERFACES][0].update(
            {
                Interface.TYPE: InterfaceType.ETHERNET,
                Ethernet.CONFIG_SUBTREE: {Ethernet.AUTO_NEGOTIATION: False},
            }
        )
        del default_data[INTERFACES][0]["link-speed"]

        libnmstate.validator.schema_validate(default_data)

    @pytest.mark.parametrize("valid_values", [0, 150, 256])
    def test_valid_with_sriov_total_vfs(self, default_data, valid_values):
        default_data[Interface.KEY][0].update(
            {
                Interface.TYPE: InterfaceType.ETHERNET,
                Ethernet.CONFIG_SUBTREE: {
                    Ethernet.SRIOV_SUBTREE: {
                        Ethernet.SRIOV.TOTAL_VFS: valid_values
                    },
                },
            }
        )
        libnmstate.validator.schema_validate(default_data)

    @pytest.mark.parametrize("invalid_values", [-50, -1])
    def test_over_maximum_total_vfs_is_invalid(
        self, default_data, invalid_values
    ):
        default_data[Interface.KEY][0].update(
            {
                Interface.TYPE: InterfaceType.ETHERNET,
                Ethernet.CONFIG_SUBTREE: {
                    Ethernet.SRIOV_SUBTREE: {
                        Ethernet.SRIOV.TOTAL_VFS: invalid_values
                    },
                },
            }
        )

        with pytest.raises(js.ValidationError, match=str(invalid_values)):
            libnmstate.validator.schema_validate(default_data)

    @pytest.mark.parametrize("vf_id", [-50, -1])
    def test_invalid_vf_ids(self, default_data, vf_id):
        default_data[Interface.KEY][0].update(
            {
                Interface.TYPE: InterfaceType.ETHERNET,
                Ethernet.CONFIG_SUBTREE: {
                    Ethernet.SRIOV_SUBTREE: {
                        Ethernet.SRIOV.TOTAL_VFS: 2,
                        Ethernet.SRIOV.VFS_SUBTREE: [
                            {Ethernet.SRIOV.VFS.ID: vf_id}
                        ],
                    },
                },
            }
        )

        with pytest.raises(js.ValidationError, match=str(vf_id)):
            libnmstate.validator.schema_validate(default_data)

    @pytest.mark.parametrize("vf_id", [0, 1, 20])
    def test_valid_vf_ids(self, default_data, vf_id):
        default_data[Interface.KEY][0].update(
            {
                Interface.TYPE: InterfaceType.ETHERNET,
                Ethernet.CONFIG_SUBTREE: {
                    Ethernet.SRIOV_SUBTREE: {
                        Ethernet.SRIOV.TOTAL_VFS: 2,
                        Ethernet.SRIOV.VFS_SUBTREE: [
                            {Ethernet.SRIOV.VFS.ID: vf_id},
                        ],
                    },
                },
            }
        )
        libnmstate.validator.schema_validate(default_data)

    def test_sriov_with_empty_vf_config_is_valid(self, default_data):
        default_data[Interface.KEY][0].update(
            {
                Interface.TYPE: InterfaceType.ETHERNET,
                Ethernet.CONFIG_SUBTREE: {
                    Ethernet.SRIOV_SUBTREE: {
                        Ethernet.SRIOV.TOTAL_VFS: 1,
                        Ethernet.SRIOV.VFS_SUBTREE: [],
                    },
                },
            }
        )
        libnmstate.validator.schema_validate(default_data)

    def test_sriov_vf_config_is_valid(self, default_data):
        default_data[Interface.KEY][0].update(
            {
                Interface.TYPE: InterfaceType.ETHERNET,
                Ethernet.CONFIG_SUBTREE: {
                    Ethernet.SRIOV_SUBTREE: {
                        Ethernet.SRIOV.TOTAL_VFS: 1,
                        Ethernet.SRIOV.VFS_SUBTREE: [
                            {
                                Ethernet.SRIOV.VFS.ID: 1,
                                Ethernet.SRIOV.VFS.MAC_ADDRESS: SRIOV_MAC,
                                Ethernet.SRIOV.VFS.SPOOF_CHECK: True,
                                Ethernet.SRIOV.VFS.TRUST: False,
                                Ethernet.SRIOV.VFS.MIN_TX_RATE: 1000,
                                Ethernet.SRIOV.VFS.MAX_TX_RATE: 2000,
                            }
                        ],
                    },
                },
            }
        )
        libnmstate.validator.schema_validate(default_data)


class TestIfaceTypeVxlan:
    def test_bad_id_type_is_invalid(self, default_data):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: VXLAN0,
                Interface.TYPE: VXLAN.TYPE,
                VXLAN.CONFIG_SUBTREE: {
                    VXLAN.ID: "badtype",
                    VXLAN.BASE_IFACE: "eth1",
                    VXLAN.REMOTE: "192.168.3.3",
                },
            }
        )

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)

    def test_bad_id_range_is_invalid(self, default_data):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: VXLAN0,
                Interface.TYPE: VXLAN.TYPE,
                VXLAN.CONFIG_SUBTREE: {
                    VXLAN.ID: 16777216,
                    VXLAN.BASE_IFACE: "eth1",
                    VXLAN.REMOTE: "192.168.3.3",
                },
            }
        )

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)

    def test_no_config_is_valid(self, default_data):
        default_data[Interface.KEY].append(
            {Interface.NAME: VXLAN0, Interface.TYPE: VXLAN.TYPE}
        )

        libnmstate.validator.schema_validate(default_data)


class TestIfaceTypeTeam:
    def test_valid_team_without_options(self, default_data):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: TEAM0,
                Interface.TYPE: Team.TYPE,
                Team.CONFIG_SUBTREE: {},
            }
        )

        libnmstate.validator.schema_validate(default_data)

    def test_valid_team_with_ports(self, default_data):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: TEAM0,
                Interface.TYPE: Team.TYPE,
                Team.CONFIG_SUBTREE: {
                    Team.PORT_SUBTREE: [{Team.Port.NAME: "eth1"}]
                },
            }
        )

        libnmstate.validator.schema_validate(default_data)

    def test_valid_team_with_runner(self, default_data):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: TEAM0,
                Interface.TYPE: Team.TYPE,
                Team.CONFIG_SUBTREE: {
                    Team.PORT_SUBTREE: [],
                    Team.RUNNER_SUBTREE: {
                        Team.Runner.NAME: Team.Runner.RunnerMode.LOAD_BALANCE
                    },
                },
            }
        )

        libnmstate.validator.schema_validate(default_data)

    def test_valid_team_with_ports_and_runner(self, default_data):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: TEAM0,
                Interface.TYPE: Team.TYPE,
                Team.CONFIG_SUBTREE: {
                    Team.PORT_SUBTREE: [{Team.Port.NAME: "eth1"}],
                    Team.RUNNER_SUBTREE: {
                        Team.Runner.NAME: Team.Runner.RunnerMode.LOAD_BALANCE
                    },
                },
            }
        )

        libnmstate.validator.schema_validate(default_data)


class TestIfaceTypeMacVlan:
    @pytest.mark.parametrize(
        "mode",
        [
            MacVlan.Mode.VEPA,
            MacVlan.Mode.BRIDGE,
            MacVlan.Mode.PRIVATE,
            MacVlan.Mode.PASSTHRU,
            MacVlan.Mode.SOURCE,
        ],
    )
    def test_valid_mac_vlan_modes(self, default_data, mode):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: "macvlan0",
                Interface.TYPE: MacVlan.TYPE,
                MacVlan.CONFIG_SUBTREE: {
                    MacVlan.BASE_IFACE: "eth1",
                    MacVlan.MODE: mode,
                    MacVlan.PROMISCUOUS: True,
                },
            }
        )
        libnmstate.validator.schema_validate(default_data)

    def test_invalid_mac_vlan_mode(self, default_data):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: "macvlan0",
                Interface.TYPE: MacVlan.TYPE,
                MacVlan.CONFIG_SUBTREE: {
                    MacVlan.BASE_IFACE: "eth1",
                    MacVlan.MODE: "wrong-mode",
                    MacVlan.PROMISCUOUS: True,
                },
            }
        )

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)


class TestIfaceTypeMacVtap:
    @pytest.mark.parametrize(
        "mode",
        [
            MacVtap.Mode.VEPA,
            MacVtap.Mode.BRIDGE,
            MacVtap.Mode.PRIVATE,
            MacVtap.Mode.PASSTHRU,
            MacVtap.Mode.SOURCE,
        ],
    )
    def test_valid_mac_vtap_modes(self, default_data, mode):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: "macvtap0",
                Interface.TYPE: MacVtap.TYPE,
                MacVtap.CONFIG_SUBTREE: {
                    MacVtap.BASE_IFACE: "eth1",
                    MacVtap.MODE: mode,
                    MacVtap.PROMISCUOUS: True,
                },
            }
        )
        libnmstate.validator.schema_validate(default_data)

    def test_invalid_mac_vtap_mode(self, default_data):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: "macvtap0",
                Interface.TYPE: MacVtap.TYPE,
                MacVtap.CONFIG_SUBTREE: {
                    MacVtap.BASE_IFACE: "eth1",
                    MacVtap.MODE: "wrong-mode",
                    MacVtap.PROMISCUOUS: True,
                },
            }
        )

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)


class TestRoutes:
    def test_valid_state_absent(self, default_data):
        default_data[ROUTES][Route.CONFIG][0][Route.STATE] = "absent"
        libnmstate.validator.schema_validate(default_data)

    def test_invalid_state(self, default_data):
        default_data[ROUTES][Route.CONFIG][0][Route.STATE] = "bad-state"

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)


class TestLinuxBridgeVlanFiltering:
    @pytest.mark.parametrize(
        "port_type",
        argvalues=[LB.Port.Vlan.Mode.TRUNK, LB.Port.Vlan.Mode.ACCESS],
    )
    def test_vlan_port_types(self, default_data, bridge_state, port_type):
        valid_port_type = generate_vlan_filtering_config(port_type)
        the_port = bridge_state[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][0]
        the_port.update(valid_port_type)
        default_data[Interface.KEY].append(bridge_state)

        libnmstate.validator.schema_validate(default_data)

    def test_invalid_vlan_port_type(self, default_data, bridge_state):
        invalid_port_type = generate_vlan_filtering_config("fake-type")
        the_port = bridge_state[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][0]
        the_port.update(invalid_port_type)
        default_data[Interface.KEY].append(bridge_state)

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)

    def test_access_port_accepted(self, default_data, bridge_state):
        vlan_access_port_state = generate_vlan_filtering_config(
            LB.Port.Vlan.Mode.ACCESS, tag=101
        )
        the_port = bridge_state[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][0]
        the_port.update(vlan_access_port_state)
        default_data[Interface.KEY].append(bridge_state)

        libnmstate.validator.schema_validate(default_data)

    def test_wrong_access_port_tag_type(self, default_data, bridge_state):
        invalid_access_port_tag_type = generate_vlan_filtering_config(
            LB.Port.Vlan.Mode.ACCESS, tag="holy-guacamole!"
        )
        the_port = bridge_state[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][0]
        the_port.update(invalid_access_port_tag_type)
        default_data[Interface.KEY].append(bridge_state)

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)

    def test_wrong_access_tag_range(self, default_data, bridge_state):
        invalid_vlan_id_range = generate_vlan_filtering_config(
            LB.Port.Vlan.Mode.ACCESS, tag=48000
        )
        the_port = bridge_state[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][0]
        the_port.update(invalid_vlan_id_range)
        default_data[Interface.KEY].append(bridge_state)

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)

    @pytest.mark.parametrize(
        "is_native_vlan", argvalues=[True, False], ids=["native", "not-native"]
    )
    def test_trunk_port_native_vlan(
        self, default_data, bridge_state, is_native_vlan
    ):
        vlan_access_port_state = generate_vlan_filtering_config(
            LB.Port.Vlan.Mode.TRUNK,
            tag=101 if is_native_vlan else None,
            native_vlan=is_native_vlan,
        )
        the_port = bridge_state[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][0]
        the_port.update(vlan_access_port_state)
        default_data[Interface.KEY].append(bridge_state)

        libnmstate.validator.schema_validate(default_data)

    def test_trunk_ports(self, default_data, bridge_state):
        trunk_tags = generate_vlan_id_config(101, 102, 103)
        trunk_tags.append(generate_vlan_id_range_config(500, 1000))
        vlan_trunk_tags_port_state = generate_vlan_filtering_config(
            LB.Port.Vlan.Mode.TRUNK, trunk_tags=trunk_tags
        )
        the_port = bridge_state[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][0]
        the_port.update(vlan_trunk_tags_port_state)
        default_data[Interface.KEY].append(bridge_state)

        libnmstate.validator.schema_validate(default_data)

    def test_invalid_trunk_port_vlan_range(self, default_data, bridge_state):
        invalid_port_vlan_configuration = generate_vlan_filtering_config(
            LB.Port.Vlan.Mode.TRUNK,
            trunk_tags=[generate_vlan_id_range_config(100, 5000)],
        )
        the_port = bridge_state[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][0]
        the_port.update(invalid_port_vlan_configuration)
        default_data[Interface.KEY].append(bridge_state)

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)


class TestOvsBridgeVlan:
    @pytest.mark.parametrize(
        "vlan_mode",
        argvalues=[
            OVSBridge.Port.Vlan.Mode.TRUNK,
            OVSBridge.Port.Vlan.Mode.ACCESS,
        ],
    )
    def test_vlan_port_modes(self, default_data, ovs_bridge_state, vlan_mode):
        valid_vlan_mode = self._generate_vlan_config(vlan_mode)
        bridge_state_config = ovs_bridge_state[OVSBridge.CONFIG_SUBTREE]
        the_port = bridge_state_config[OVSBridge.PORT_SUBTREE][0]
        the_port.update(valid_vlan_mode)
        default_data[Interface.KEY].append(ovs_bridge_state)

        libnmstate.validator.schema_validate(default_data)

    def test_invalid_vlan_port_mode(self, default_data, ovs_bridge_state):
        invalid_vlan_mode = self._generate_vlan_config("fake-mode")
        bridge_state_config = ovs_bridge_state[OVSBridge.CONFIG_SUBTREE]
        the_port = bridge_state_config[OVSBridge.PORT_SUBTREE][0]
        the_port.update(invalid_vlan_mode)
        default_data[Interface.KEY].append(ovs_bridge_state)

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)

    def test_access_port_accepted(self, default_data, ovs_bridge_state):
        vlan_access_port_state = self._generate_vlan_config(
            OVSBridge.Port.Vlan.Mode.ACCESS, access_tag=101
        )
        bridge_state_config = ovs_bridge_state[OVSBridge.CONFIG_SUBTREE]
        the_port = bridge_state_config[OVSBridge.PORT_SUBTREE][0]
        the_port.update(vlan_access_port_state)
        default_data[Interface.KEY].append(ovs_bridge_state)

        libnmstate.validator.schema_validate(default_data)

    def test_wrong_access_port_tag_mode(self, default_data, ovs_bridge_state):
        invalid_access_port_tag_mode = self._generate_vlan_config(
            OVSBridge.Port.Vlan.Mode.ACCESS, access_tag="holy-guacamole!"
        )
        bridge_state_config = ovs_bridge_state[OVSBridge.CONFIG_SUBTREE]
        the_port = bridge_state_config[OVSBridge.PORT_SUBTREE][0]
        the_port.update(invalid_access_port_tag_mode)
        default_data[Interface.KEY].append(ovs_bridge_state)

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)

    def test_wrong_access_tag_range(self, default_data, ovs_bridge_state):
        invalid_vlan_id_range = self._generate_vlan_config(
            OVSBridge.Port.Vlan.Mode.ACCESS, access_tag=48000
        )
        bridge_state_config = ovs_bridge_state[OVSBridge.CONFIG_SUBTREE]
        the_port = bridge_state_config[OVSBridge.PORT_SUBTREE][0]
        the_port.update(invalid_vlan_id_range)
        default_data[Interface.KEY].append(ovs_bridge_state)

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)

    @pytest.mark.parametrize(
        "is_native_vlan", argvalues=[True, False], ids=["native", "not-native"]
    )
    def test_trunk_port_native_vlan(
        self, default_data, ovs_bridge_state, is_native_vlan
    ):
        vlan_access_port_state = self._generate_vlan_config(
            OVSBridge.Port.Vlan.Mode.TRUNK,
            access_tag=101 if is_native_vlan else None,
            native_vlan=is_native_vlan,
        )
        bridge_state_config = ovs_bridge_state[OVSBridge.CONFIG_SUBTREE]
        the_port = bridge_state_config[OVSBridge.PORT_SUBTREE][0]
        the_port.update(vlan_access_port_state)
        default_data[Interface.KEY].append(ovs_bridge_state)

        libnmstate.validator.schema_validate(default_data)

    def test_trunk_ports(self, default_data, ovs_bridge_state):
        trunk_tags = self._generate_vlan_id_config(101, 102, 103)
        trunk_tags.append(self._generate_vlan_id_range_config(500, 1000))
        vlan_trunk_tags_port_state = self._generate_vlan_config(
            OVSBridge.Port.Vlan.Mode.TRUNK, trunk_tags=trunk_tags
        )
        bridge_state_config = ovs_bridge_state[OVSBridge.CONFIG_SUBTREE]
        the_port = bridge_state_config[OVSBridge.PORT_SUBTREE][0]
        the_port.update(vlan_trunk_tags_port_state)
        default_data[Interface.KEY].append(ovs_bridge_state)

        libnmstate.validator.schema_validate(default_data)

    def test_invalid_trunk_port_vlan_range(
        self, default_data, ovs_bridge_state
    ):
        invalid_port_vlan_configuration = self._generate_vlan_config(
            OVSBridge.Port.Vlan.Mode.TRUNK,
            trunk_tags=[self._generate_vlan_id_range_config(100, 5000)],
        )
        bridge_state_config = ovs_bridge_state[OVSBridge.CONFIG_SUBTREE]
        the_port = bridge_state_config[OVSBridge.PORT_SUBTREE][0]
        the_port.update(invalid_port_vlan_configuration)
        default_data[Interface.KEY].append(ovs_bridge_state)

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)

    @staticmethod
    def _generate_vlan_config(
        vlan_mode, trunk_tags=None, access_tag=None, native_vlan=None
    ):
        vlan_state = {
            OVSBridge.Port.Vlan.MODE: vlan_mode,
            OVSBridge.Port.Vlan.TRUNK_TAGS: trunk_tags or [],
        }

        if access_tag:
            vlan_state[OVSBridge.Port.Vlan.TAG] = access_tag
        if native_vlan:
            enable_native = OVSBridge.Port.Vlan.ENABLE_NATIVE
            vlan_state[enable_native] = native_vlan

        return {OVSBridge.Port.VLAN_SUBTREE: vlan_state}

    @staticmethod
    def _generate_vlan_id_config(*vlan_ids):
        return [
            {OVSBridge.Port.Vlan.TrunkTags.ID: vlan_id} for vlan_id in vlan_ids
        ]

    @staticmethod
    def _generate_vlan_id_range_config(min_vlan_id, max_vlan_id):
        return {
            OVSBridge.Port.Vlan.TrunkTags.ID_RANGE: {
                OVSBridge.Port.Vlan.TrunkTags.MIN_RANGE: min_vlan_id,
                OVSBridge.Port.Vlan.TrunkTags.MAX_RANGE: max_vlan_id,
            }
        }


class TestRouteRules:
    def test_non_interger_route_table(self, default_data):
        route_rules = default_data[RouteRule.KEY][RouteRule.CONFIG]
        route_rules[0][RouteRule.ROUTE_TABLE] = "main"

        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)

    def test_route_rule_state_absent(self, default_data):
        route_rules = default_data[RouteRule.KEY]
        route_rules[RouteRule.CONFIG] = [
            {RouteRule.STATE: RouteRule.STATE_ABSENT}
        ]

        libnmstate.validator.schema_validate(default_data)


class TestOVSBridgeLinkAggregation:
    def test_valid_link_aggregation_port(self, default_data):
        link_aggregation_port = {
            OVSBridge.Port.NAME: "bond",
            OVSBridge.Port.LINK_AGGREGATION_SUBTREE: {
                OVSBridge.Port.LinkAggregation.MODE: "bond-mode",
                OVSBridge.Port.LinkAggregation.PORT_SUBTREE: [
                    {OVSBridge.Port.LinkAggregation.Port.NAME: "iface1"},
                    {OVSBridge.Port.LinkAggregation.Port.NAME: "iface2"},
                ],
            },
        }
        default_data[Interface.KEY].append(
            {
                Interface.NAME: "bridge",
                Interface.TYPE: InterfaceType.OVS_BRIDGE,
                OVSBridge.CONFIG_SUBTREE: {
                    OVSBridge.PORT_SUBTREE: [link_aggregation_port]
                },
            }
        )

        libnmstate.validator.schema_validate(default_data)


def generate_vlan_filtering_config(
    port_type, trunk_tags=None, tag=None, native_vlan=None
):
    vlan_filtering_state = {
        LB.Port.Vlan.MODE: port_type,
        LB.Port.Vlan.TRUNK_TAGS: trunk_tags or [],
    }

    if tag:
        vlan_filtering_state[LB.Port.Vlan.TAG] = tag
    if native_vlan is not None:
        vlan_filtering_state[LB.Port.Vlan.ENABLE_NATIVE] = native_vlan

    return {LB.Port.VLAN_SUBTREE: vlan_filtering_state}


def generate_vlan_id_config(*vlan_ids):
    return [{LB.Port.Vlan.TrunkTags.ID: vlan_id} for vlan_id in vlan_ids]


def generate_vlan_id_range_config(min_vlan_id, max_vlan_id):
    return {
        LB.Port.Vlan.TrunkTags.ID_RANGE: {
            LB.Port.Vlan.TrunkTags.MIN_RANGE: min_vlan_id,
            LB.Port.Vlan.TrunkTags.MAX_RANGE: max_vlan_id,
        }
    }


class TestOVSInterface:
    def test_valid_ovs_interface_with_peer(self, default_data):
        ovs_patch0 = {OVSInterface.Patch.PEER: "ovs1"}
        ovs_patch1 = {OVSInterface.Patch.PEER: "ovs0"}
        default_data[Interface.KEY].append(
            {
                Interface.NAME: "ovs0",
                Interface.TYPE: InterfaceType.OVS_INTERFACE,
                OVSInterface.PATCH_CONFIG_SUBTREE: ovs_patch0,
            }
        )
        default_data[Interface.KEY].append(
            {
                Interface.NAME: "ovs1",
                Interface.TYPE: InterfaceType.OVS_INTERFACE,
                OVSInterface.PATCH_CONFIG_SUBTREE: ovs_patch1,
            }
        )

        libnmstate.validator.schema_validate(default_data)

    def test_valid_ovs_interface_without_peer(self, default_data):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: "ovs0",
                Interface.TYPE: InterfaceType.OVS_INTERFACE,
            }
        )

        libnmstate.validator.schema_validate(default_data)


class TestIfaceTypeVrf:
    def _gen_base_vrf_iface_info(self):
        return {
            Interface.NAME: "vrf0",
            Interface.TYPE: InterfaceType.VRF,
            VRF.CONFIG_SUBTREE: {
                VRF.PORT_SUBTREE: ["foo1", "foo2"],
                VRF.ROUTE_TABLE_ID: 100,
            },
        }

    def test_valid_vrf_interface_with_ports(self):
        libnmstate.validator.schema_validate(
            {Interface.KEY: [self._gen_base_vrf_iface_info()]}
        )


class TestInfiniBand:
    def test_valid_base_interface_full_info(self, default_data):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: "ib0",
                Interface.TYPE: InterfaceType.INFINIBAND,
                InfiniBand.CONFIG_SUBTREE: {
                    InfiniBand.MODE: InfiniBand.Mode.DATAGRAM,
                    InfiniBand.PKEY: InfiniBand.DEFAULT_PKEY,
                    InfiniBand.BASE_IFACE: "",
                },
            }
        )
        libnmstate.validator.schema_validate(default_data)

    def test_valid_base_interface_minimum_info(self, default_data):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: "ib0",
                Interface.TYPE: InterfaceType.INFINIBAND,
                InfiniBand.CONFIG_SUBTREE: {
                    InfiniBand.MODE: InfiniBand.Mode.DATAGRAM,
                },
            }
        )
        libnmstate.validator.schema_validate(default_data)

    def test_valid_pkey_interface_using_integer_pkey(self, default_data):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: "ib0.80ff",
                Interface.TYPE: InterfaceType.INFINIBAND,
                InfiniBand.CONFIG_SUBTREE: {
                    InfiniBand.MODE: InfiniBand.Mode.DATAGRAM,
                    InfiniBand.PKEY: 0x80FF,
                    InfiniBand.BASE_IFACE: "ib0",
                },
            }
        )
        libnmstate.validator.schema_validate(default_data)

    def test_valid_pkey_interface_using_string_pkey(self, default_data):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: "ib0.80ff",
                Interface.TYPE: InterfaceType.INFINIBAND,
                InfiniBand.CONFIG_SUBTREE: {
                    InfiniBand.MODE: InfiniBand.Mode.DATAGRAM,
                    InfiniBand.PKEY: "0x80FF",
                    InfiniBand.BASE_IFACE: "ib0",
                },
            }
        )
        libnmstate.validator.schema_validate(default_data)

    def test_valid_pkey_interface_using_invalid_mode(self, default_data):
        default_data[Interface.KEY].append(
            {
                Interface.NAME: "ib0.80ff",
                Interface.TYPE: InterfaceType.INFINIBAND,
                InfiniBand.CONFIG_SUBTREE: {
                    InfiniBand.MODE: "invalid",
                    InfiniBand.PKEY: "0x80FF",
                    InfiniBand.BASE_IFACE: "ib0",
                },
            }
        )
        with pytest.raises(js.ValidationError):
            libnmstate.validator.schema_validate(default_data)



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import time

import pytest

import libnmstate

from libnmstate.error import NmstateVerificationError
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState
from libnmstate.schema import VXLAN

from .testlib import assertlib
from .testlib import statelib
from .testlib.bondlib import bond_interface
from .testlib.cmdlib import RC_SUCCESS
from .testlib.cmdlib import exec_cmd
from .testlib.cmdlib import format_exec_cmd_result
from .testlib.vxlan import VxlanState
from .testlib.vxlan import vxlan_interfaces
from .testlib.vxlan import vxlans_absent
from .testlib.vxlan import vxlans_down
from .testlib.vxlan import vxlans_up

VXLAN1_ID = 201
VXLAN2_ID = 202

VXLAN1_IFNAME = f"eth1.{VXLAN1_ID}"


def test_add_and_remove_vxlan(eth1_up):
    ifname = eth1_up[Interface.KEY][0][Interface.NAME]
    with vxlan_interfaces(
        VxlanState(id=VXLAN1_ID, base_if=ifname, remote="192.168.100.1")
    ) as desired_state:
        assertlib.assert_state(desired_state)

    vxlan1_ifname = desired_state[Interface.KEY][0][Interface.NAME]
    assertlib.assert_absent(vxlan1_ifname)


@pytest.mark.tier1
def test_add_and_remove_two_vxlans_on_same_iface(eth1_up):
    ifname = eth1_up[Interface.KEY][0][Interface.NAME]
    with vxlan_interfaces(
        VxlanState(id=VXLAN1_ID, base_if=ifname, remote="192.168.100.1"),
        VxlanState(id=VXLAN2_ID, base_if=ifname, remote="192.168.100.2"),
    ) as desired_state:
        assertlib.assert_state(desired_state)

    vxlan_interfaces_name = (
        i[Interface.NAME] for i in desired_state[Interface.KEY]
    )
    assertlib.assert_absent(vxlan_interfaces_name)


@pytest.mark.tier1
def test_rollback_for_vxlans(eth1_up):
    ifname = eth1_up[Interface.KEY][0][Interface.NAME]
    current_state = libnmstate.show()
    desired_state = vxlans_up(
        [
            VxlanState(id=VXLAN1_ID, base_if=ifname, remote="192.168.100.1"),
            VxlanState(id=VXLAN2_ID, base_if=ifname, remote="192.168.100.2"),
        ]
    )
    desired_state[Interface.KEY][1]["invalid_key"] = "foo"
    with pytest.raises(NmstateVerificationError):
        libnmstate.apply(desired_state)

    time.sleep(5)  # Give some time for NetworkManager to rollback
    current_state_after_apply = libnmstate.show()
    assert current_state == current_state_after_apply


def test_set_vxlan_iface_down(eth1_up):
    ifname = eth1_up[Interface.KEY][0][Interface.NAME]
    vxlan = VxlanState(id=VXLAN1_ID, base_if=ifname, remote="192.168.100.1")
    with vxlan_interfaces(vxlan):
        desired_state = vxlans_down([vxlan])
        libnmstate.apply(desired_state)
        assertlib.assert_absent(vxlan.name)


def test_add_new_bond_iface_with_vxlan(eth1_up):
    eth_name = eth1_up[Interface.KEY][0][Interface.NAME]
    bond_name = "bond0"
    vxlan = VxlanState(id=VXLAN1_ID, base_if=bond_name, remote="192.168.100.2")
    with bond_interface(
        name=bond_name, port=[eth_name], extra_iface_state=None, create=False
    ) as bond_desired_state:
        with vxlan_interfaces(vxlan, create=False) as desired_state:
            desired_state[Interface.KEY].append(
                bond_desired_state[Interface.KEY][0]
            )
            libnmstate.apply(desired_state)
            assertlib.assert_state_match(desired_state)

    assertlib.assert_absent(vxlan.name)
    assertlib.assert_absent(bond_name)


def test_show_vxlan_with_no_remote(eth1_up):
    eth_name = eth1_up[Interface.KEY][0][Interface.NAME]
    vxlan = VxlanState(id=VXLAN1_ID, base_if=eth_name, remote="")
    add_vxlan_cmd = (
        f"ip link add {vxlan.name} type vxlan id {vxlan.id}"
        f" dstport {vxlan.destination_port} dev {eth_name}".split()
    )
    try:
        ret = exec_cmd(add_vxlan_cmd)
        rc, _, _ = ret
        assert rc == RC_SUCCESS, format_exec_cmd_result(ret)
        desired_state = vxlans_down([vxlan])
        assertlib.assert_state(desired_state)
    finally:
        libnmstate.apply(vxlans_absent([vxlan]))
        assertlib.assert_absent(vxlan.name)


@pytest.fixture
def vxlan1_up(eth1_up):
    with vxlan_interfaces(
        VxlanState(id=VXLAN1_ID, base_if="eth1", remote="192.168.100.1")
    ):
        yield


def test_show_saved_config_with_vxlan_down(vxlan1_up):
    running_state = statelib.show_only((VXLAN1_IFNAME,))
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: VXLAN1_IFNAME,
                    Interface.STATE: InterfaceState.DOWN,
                }
            ]
        }
    )
    saved_state = statelib.show_saved_config_only((VXLAN1_IFNAME,))
    iface_state = saved_state[Interface.KEY][0]

    assert saved_state[Interface.KEY][0][Interface.STATE] == InterfaceState.UP
    assert iface_state[VXLAN.CONFIG_SUBTREE] == {
        VXLAN.ID: VXLAN1_ID,
        VXLAN.BASE_IFACE: "eth1",
    }
    assertlib.assert_state_match_full(saved_state, running_state)



#
# Copyright (c) 2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
from contextlib import contextmanager

import libnmstate
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType
from libnmstate.schema import LinuxBridge


@contextmanager
def linux_bridge(
    name, bridge_subtree_state, extra_iface_state=None, create=True
):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: name,
                Interface.TYPE: InterfaceType.LINUX_BRIDGE,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV4: {InterfaceIPv4.ENABLED: False},
                Interface.IPV6: {InterfaceIPv6.ENABLED: False},
            }
        ]
    }
    desired_iface_state = desired_state[Interface.KEY][0]
    if bridge_subtree_state:
        desired_iface_state[LinuxBridge.CONFIG_SUBTREE] = bridge_subtree_state
    if extra_iface_state:
        desired_iface_state.update(extra_iface_state)

    if create:
        libnmstate.apply(desired_state)

    try:
        yield desired_state
    finally:
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: name,
                        Interface.TYPE: InterfaceType.LINUX_BRIDGE,
                        Interface.STATE: InterfaceState.ABSENT,
                    }
                ]
            },
            verify_change=False,
        )


def add_port_to_bridge(bridge_subtree_state, port_name, port_state=None):
    if LinuxBridge.PORT_SUBTREE not in bridge_subtree_state:
        bridge_subtree_state[LinuxBridge.PORT_SUBTREE] = []

    if port_state is None:
        port_state = {}

    port_state[LinuxBridge.Port.NAME] = port_name
    bridge_subtree_state[LinuxBridge.PORT_SUBTREE].append(port_state)

    return bridge_subtree_state


def create_bridge_subtree_state(options_state=None):
    if options_state is None:
        options_state = {
            LinuxBridge.STP_SUBTREE: {LinuxBridge.STP.ENABLED: False}
        }
    return {LinuxBridge.OPTIONS_SUBTREE: options_state}


def generate_vlan_filtering_config(
    port_type, trunk_tags=None, tag=None, native_vlan=None
):
    vlan_filtering_state = {
        LinuxBridge.Port.Vlan.MODE: port_type,
        LinuxBridge.Port.Vlan.TRUNK_TAGS: trunk_tags or [],
    }

    if tag:
        vlan_filtering_state[LinuxBridge.Port.Vlan.TAG] = tag
    if native_vlan is not None:
        vlan_filtering_state[LinuxBridge.Port.Vlan.ENABLE_NATIVE] = native_vlan

    return {LinuxBridge.Port.VLAN_SUBTREE: vlan_filtering_state}


def generate_vlan_id_config(*vlan_ids):
    return [
        {LinuxBridge.Port.Vlan.TrunkTags.ID: vlan_id} for vlan_id in vlan_ids
    ]


def generate_vlan_id_range_config(min_vlan_id, max_vlan_id):
    return {
        LinuxBridge.Port.Vlan.TrunkTags.ID_RANGE: {
            LinuxBridge.Port.Vlan.TrunkTags.MIN_RANGE: min_vlan_id,
            LinuxBridge.Port.Vlan.TrunkTags.MAX_RANGE: max_vlan_id,
        }
    }



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import time

import pytest

import libnmstate
from libnmstate.schema import Bond
from libnmstate.schema import BondMode
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState

from ..testlib import cmdlib

BOND99 = "bond99"
DUMMY1 = "dummy1"
DUMMY2 = "dummy2"


@pytest.fixture
def bond99_with_dummy_port_by_iproute():
    cmdlib.exec_cmd(f"ip link add {DUMMY1} type dummy".split(), check=True)
    cmdlib.exec_cmd(f"ip link add {DUMMY2} type dummy".split(), check=True)
    cmdlib.exec_cmd(f"ip link add {BOND99} type bond".split(), check=True)
    cmdlib.exec_cmd(
        f"ip link set {DUMMY1} master {BOND99}".split(), check=True
    )
    cmdlib.exec_cmd(
        f"ip link set {DUMMY2} master {BOND99}".split(), check=True
    )
    cmdlib.exec_cmd(f"ip link set {DUMMY1} up".split(), check=True)
    cmdlib.exec_cmd(f"ip link set {DUMMY2} up".split(), check=True)
    cmdlib.exec_cmd(f"ip link set {BOND99} up".split(), check=True)
    time.sleep(1)  # Wait NM mark them as managed
    yield
    cmdlib.exec_cmd(f"nmcli c del {BOND99}".split())
    cmdlib.exec_cmd(f"nmcli c del {DUMMY1}".split())
    cmdlib.exec_cmd(f"nmcli c del {DUMMY2}".split())
    cmdlib.exec_cmd(f"ip link del {DUMMY1}".split())
    cmdlib.exec_cmd(f"ip link del {DUMMY2}".split())
    cmdlib.exec_cmd(f"ip link del {BOND99}".split())


def test_external_managed_subordnates(bond99_with_dummy_port_by_iproute):
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: BOND99,
                    Interface.STATE: InterfaceState.UP,
                    Bond.CONFIG_SUBTREE: {
                        # Change the bond mode to force a reactivate
                        Bond.MODE: BondMode.ACTIVE_BACKUP,
                        Bond.PORT: [DUMMY1, DUMMY2],
                    },
                }
            ]
        }
    )



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from operator import itemgetter
import warnings

from libnmstate.schema import Team

from .base_iface import BaseIface


DEPRECATED_PORTS = "ports"


class TeamIface(BaseIface):
    def __init__(self, info, save_to_disk=True):
        super().__init__(info, save_to_disk)
        self._replace_deprecated_terms()

    @property
    def port(self):
        ports = self.raw.get(Team.CONFIG_SUBTREE, {}).get(
            Team.PORT_SUBTREE, []
        )
        return [p[Team.Port.NAME] for p in ports]

    @property
    def is_controller(self):
        return True

    @property
    def is_virtual(self):
        return True

    def sort_port(self):
        if self.port:
            self.raw[Team.CONFIG_SUBTREE][Team.PORT_SUBTREE].sort(
                key=itemgetter(Team.Port.NAME)
            )

    def remove_port(self, port_name):
        if self.port:
            port_config = self.raw[Team.CONFIG_SUBTREE][Team.PORT_SUBTREE]
            self.raw[Team.CONFIG_SUBTREE][Team.PORT_SUBTREE] = [
                s for s in port_config if s[Team.Port.NAME] != port_name
            ]
        self.sort_port()

    def _replace_deprecated_terms(self):
        team_cfg = self.raw.get(Team.CONFIG_SUBTREE)
        if team_cfg and team_cfg.get(DEPRECATED_PORTS):
            team_cfg[Team.PORT_SUBTREE] = team_cfg.pop(DEPRECATED_PORTS)
            warnings.warn("Using 'ports' is deprecated, use 'port' instead.")



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
from unittest import mock

import pytest

from libnmstate import nm


@pytest.fixture()
def client_mock():
    yield mock.MagicMock()


def test_list_devices(client_mock):
    nm.device.list_devices(client_mock)

    client_mock.get_devices.assert_called_once()


def test_get_device_common_info():
    dev = mock.MagicMock()

    info = nm.device.get_device_common_info(dev)

    expected_info = {
        "name": dev.get_iface.return_value,
        "type_id": dev.get_device_type.return_value,
        "type_name": dev.get_type_description.return_value,
        "state": dev.get_state.return_value,
    }
    assert expected_info == info



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
import pytest
import varlink
import threading
import subprocess
import sys
import json
import libnmstate
import os
import time

from nmstatectl.nmstate_varlink import gen_varlink_server
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6


VARLINK_INTERFACE = "io.nmstate"

APPLY_STATE = {
    Interface.KEY: [
        {
            Interface.NAME: "varlink_test",
            Interface.TYPE: InterfaceType.DUMMY,
            Interface.STATE: InterfaceState.UP,
            Interface.IPV4: {InterfaceIPv4.ENABLED: False},
            Interface.IPV6: {InterfaceIPv6.ENABLED: False},
        }
    ]
}


@pytest.fixture(
    params=["/run/nmstate"],
    ids=["unix_socket_connection"],
)
def server(request):
    server = gen_varlink_server(request.param)
    thread_server = threading.Thread(target=server.serve_forever, daemon=True)
    thread_server.start()
    yield server
    server.shutdown()
    server.server_close()


@pytest.fixture(
    params=["unix:/run/nmstate/nmstate.so"],
    ids=["systemd_unix_connection"],
)
def systemd_service(request):
    if os.system("systemctl is-active nmstate-varlink.service") != 0:
        os.system("systemctl start nmstate-varlink.service")
        time.sleep(1)
    yield request.param
    os.system("systemctl stop nmstate-varlink.service")


def _format_address(server_address):
    return f"unix:{server_address}"


def test_varlink_show(server):
    lib_state = libnmstate.show()
    with varlink.Client(_format_address(server.server_address)).open(
        VARLINK_INTERFACE, namespaced=False
    ) as con:
        varlink_state = con._call("Show")
        assert varlink_state["state"] == lib_state


def test_varlink_show_running_config(server):
    lib_state = libnmstate.show_running_config()
    with varlink.Client(_format_address(server.server_address)).open(
        VARLINK_INTERFACE, namespaced=False
    ) as con:
        varlink_state = con._call("ShowRunningConfig")
        assert varlink_state["state"] == lib_state


def test_varlink_show_saved_config(server):
    lib_state = libnmstate.show_saved_config()
    with varlink.Client(_format_address(server.server_address)).open(
        VARLINK_INTERFACE, namespaced=False
    ) as con:
        varlink_state = con._call("ShowSavedConfig")
        assert varlink_state["state"] == lib_state


def test_varlink_apply_state(server):
    with varlink.Client(_format_address(server.server_address)).open(
        VARLINK_INTERFACE, namespaced=False
    ) as con:
        con._call("Apply", {"desired_state": APPLY_STATE})
        lib_state = libnmstate.show()
        assert any(
            interface
            for interface in lib_state[Interface.KEY]
            if interface[Interface.NAME]
            == APPLY_STATE[Interface.KEY][0][Interface.NAME]
        )
        APPLY_STATE[Interface.KEY][0][Interface.STATE] = InterfaceState.DOWN
        libnmstate.apply(APPLY_STATE)


def test_varlink_apply_with_none_state(server):
    apply_state = {"desired_state": {}}
    with varlink.Client(_format_address(server.server_address)).open(
        VARLINK_INTERFACE, namespaced=False
    ) as con:
        with pytest.raises(varlink.InvalidParameter):
            con._call("Apply", apply_state)


def test_varlink_commit_error(server):
    with varlink.Client(_format_address(server.server_address)).open(
        VARLINK_INTERFACE, namespaced=False
    ) as con:
        with pytest.raises(varlink.VarlinkError):
            con._call("Commit", None)


def test_varlink_rollback_error(server):
    with varlink.Client(_format_address(server.server_address)).open(
        VARLINK_INTERFACE, namespaced=False
    ) as con:
        with pytest.raises(varlink.VarlinkError):
            con._call("Rollback", None)


def test_varlink_activation_mode():
    lib_state = libnmstate.show()
    command = (
        "varlink --activate='nmstatectl varlink /run/nmstate' "
        + "call io.nmstate.Show"
    )
    output = subprocess.check_output(command, shell=True).decode(
        sys.stdout.encoding
    )
    parsed_out = json.loads(output)
    assert parsed_out["state"] == lib_state


def test_systemd_varlink_show(systemd_service):
    lib_state = libnmstate.show()
    with varlink.Client(systemd_service).open(
        VARLINK_INTERFACE, namespaced=False
    ) as con:
        varlink_state = con._call("Show")
        assert varlink_state["state"] == lib_state


def test_systemd_varlink_apply_state(systemd_service):
    APPLY_STATE[Interface.KEY][0][Interface.STATE] = InterfaceState.UP
    with varlink.Client(systemd_service).open(
        VARLINK_INTERFACE, namespaced=False
    ) as con:
        con._call("Apply", {"desired_state": APPLY_STATE})
        lib_state = libnmstate.show()
        assert any(
            interface
            for interface in lib_state[Interface.KEY]
            if interface[Interface.NAME]
            == APPLY_STATE[Interface.KEY][0][Interface.NAME]
        )
        APPLY_STATE[Interface.KEY][0][Interface.STATE] = InterfaceState.DOWN
        libnmstate.apply(APPLY_STATE)



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#


from .cmdlib import exec_cmd


def create_veth_pair(nic, nic_peer, peer_ns):
    """
    Create a veth pair and place the {peer} into {peer_ns} namespace.
    The {nic} will be marked as managed by NetworkManager
    """
    exec_cmd(
        f"ip link add {nic} type veth peer name {nic_peer}".split(),
        check=True,
    )
    exec_cmd(f"ip netns add {peer_ns}".split(), check=True)
    exec_cmd(f"ip link set {nic_peer} netns {peer_ns}".split(), check=True)
    exec_cmd(f"ip link set {nic} up".split(), check=True)
    exec_cmd(
        f"ip netns exec {peer_ns} ip link set {nic_peer} up".split(),
        check=True,
    )
    exec_cmd(f"nmcli device set {nic} managed yes".split(), check=True)


def remove_veth_pair(nic, peer_ns):
    exec_cmd(f"ip link del {nic}".split())
    exec_cmd(f"ip netns del {peer_ns}".split())



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager
import time

import pytest

import libnmstate
from libnmstate.error import NmstateVerificationError
from libnmstate.schema import VLAN
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType

from .testlib import assertlib
from .testlib import statelib
from .testlib.assertlib import assert_mac_address
from .testlib.vlan import vlan_interface

VLAN_IFNAME = "eth1.101"
VLAN2_IFNAME = "eth1.102"


@pytest.mark.tier1
def test_add_and_remove_vlan(eth1_up):
    with vlan_interface(
        VLAN_IFNAME, 101, eth1_up[Interface.KEY][0][Interface.NAME]
    ) as desired_state:
        assertlib.assert_state(desired_state)

    current_state = statelib.show_only((VLAN_IFNAME,))
    assert not current_state[Interface.KEY]


@pytest.fixture
def vlan_on_eth1(eth1_up):
    with vlan_interface(
        VLAN_IFNAME, 101, eth1_up[Interface.KEY][0][Interface.NAME]
    ) as desired_state:
        base_iface_name = desired_state[Interface.KEY][0][VLAN.CONFIG_SUBTREE][
            VLAN.BASE_IFACE
        ]
        iface_states = statelib.show_only((base_iface_name, VLAN_IFNAME))
        yield iface_states


@pytest.mark.tier1
def test_vlan_iface_uses_the_mac_of_base_iface(vlan_on_eth1):
    assert_mac_address(vlan_on_eth1)


def test_add_and_remove_two_vlans_on_same_iface(eth1_up):
    with two_vlans_on_eth1() as desired_state:
        assertlib.assert_state(desired_state)

    vlan_interfaces = [i[Interface.NAME] for i in desired_state[Interface.KEY]]
    current_state = statelib.show_only(vlan_interfaces)
    assert not current_state[Interface.KEY]


@pytest.mark.tier1
def test_two_vlans_on_eth1_change_mtu(eth1_up):
    with two_vlans_on_eth1() as desired_state:
        eth1_state = eth1_up[Interface.KEY][0]
        desired_state[Interface.KEY].append(eth1_state)
        for iface in desired_state[Interface.KEY]:
            iface[Interface.MTU] = 2000
        libnmstate.apply(desired_state)

        eth1_102_state = next(
            ifstate
            for ifstate in desired_state[Interface.KEY]
            if ifstate[Interface.NAME] == VLAN2_IFNAME
        )
        eth1_state[Interface.MTU] = 2200
        eth1_102_state[Interface.MTU] = 2200
        libnmstate.apply(desired_state)

        eth1_vlan_iface_cstate = statelib.show_only((VLAN_IFNAME,))
        assert eth1_vlan_iface_cstate[Interface.KEY][0][Interface.MTU] == 2000


@pytest.mark.tier1
def test_two_vlans_on_eth1_change_base_iface_mtu(eth1_up):
    with two_vlans_on_eth1() as desired_state:
        eth1_state = eth1_up[Interface.KEY][0]
        desired_state[Interface.KEY].append(eth1_state)
        for iface in desired_state[Interface.KEY]:
            iface[Interface.MTU] = 2000
        libnmstate.apply(desired_state)

        eth1_state[Interface.MTU] = 2200
        libnmstate.apply(eth1_up)
        eth1_vlan_iface_cstate = statelib.show_only((VLAN_IFNAME,))
        assert eth1_vlan_iface_cstate[Interface.KEY][0][Interface.MTU] == 2000


@pytest.mark.tier1
def test_two_vlans_on_eth1_change_mtu_rollback(eth1_up):
    with two_vlans_on_eth1() as desired_state:
        eth1_state = eth1_up[Interface.KEY][0]
        desired_state[Interface.KEY].append(eth1_state)
        for iface in desired_state[Interface.KEY]:
            iface[Interface.MTU] = 2000
        libnmstate.apply(desired_state)

        for iface in desired_state[Interface.KEY]:
            iface[Interface.MTU] = 2200
        libnmstate.apply(desired_state, commit=False)
        libnmstate.rollback()

        time.sleep(5)  # Give some time for NetworkManager to rollback

        eth1_vlan_iface_cstate = statelib.show_only((VLAN_IFNAME,))
        assert eth1_vlan_iface_cstate[Interface.KEY][0][Interface.MTU] == 2000


def test_rollback_for_vlans(eth1_up):
    current_state = libnmstate.show()
    desired_state = create_two_vlans_state()

    desired_state[Interface.KEY][1]["invalid_key"] = "foo"
    with pytest.raises(NmstateVerificationError):
        libnmstate.apply(desired_state)

    time.sleep(5)  # Give some time for NetworkManager to rollback
    current_state_after_apply = libnmstate.show()
    assert current_state == current_state_after_apply


def test_set_vlan_iface_down(eth1_up):
    with vlan_interface(
        VLAN_IFNAME, 101, eth1_up[Interface.KEY][0][Interface.NAME]
    ):
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: VLAN_IFNAME,
                        Interface.TYPE: InterfaceType.VLAN,
                        Interface.STATE: InterfaceState.DOWN,
                    }
                ]
            }
        )

        assertlib.assert_absent(VLAN_IFNAME)


def test_add_new_base_iface_with_vlan():
    iface_base = "dummy00"
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: "dummy00.101",
                Interface.TYPE: InterfaceType.VLAN,
                Interface.STATE: InterfaceState.UP,
                VLAN.CONFIG_SUBTREE: {
                    VLAN.ID: 101,
                    VLAN.BASE_IFACE: iface_base,
                },
            },
            {
                Interface.NAME: iface_base,
                Interface.TYPE: InterfaceType.DUMMY,
                Interface.STATE: InterfaceState.UP,
            },
        ]
    }
    try:
        libnmstate.apply(desired_state)
    finally:
        for ifstate in desired_state[Interface.KEY]:
            ifstate[Interface.STATE] = InterfaceState.ABSENT
        libnmstate.apply(desired_state, verify_change=False)


@contextmanager
def two_vlans_on_eth1():
    desired_state = create_two_vlans_state()
    libnmstate.apply(desired_state)
    try:
        yield desired_state
    finally:
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: VLAN_IFNAME,
                        Interface.TYPE: InterfaceType.VLAN,
                        Interface.STATE: InterfaceState.ABSENT,
                    },
                    {
                        Interface.NAME: VLAN2_IFNAME,
                        Interface.TYPE: InterfaceType.VLAN,
                        Interface.STATE: InterfaceState.ABSENT,
                    },
                ]
            }
        )


def create_two_vlans_state():
    return {
        Interface.KEY: [
            {
                Interface.NAME: VLAN_IFNAME,
                Interface.TYPE: InterfaceType.VLAN,
                Interface.STATE: InterfaceState.UP,
                VLAN.CONFIG_SUBTREE: {VLAN.ID: 101, VLAN.BASE_IFACE: "eth1"},
            },
            {
                Interface.NAME: VLAN2_IFNAME,
                Interface.TYPE: InterfaceType.VLAN,
                Interface.STATE: InterfaceState.UP,
                VLAN.CONFIG_SUBTREE: {VLAN.ID: 102, VLAN.BASE_IFACE: "eth1"},
            },
        ]
    }


@pytest.fixture
def vlan_up(eth1_up):
    with vlan_interface(VLAN_IFNAME, 101, "eth1"):
        yield


def test_show_saved_config_with_vlan_down(vlan_up):
    running_state = statelib.show_only((VLAN_IFNAME,))
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: VLAN_IFNAME,
                    Interface.STATE: InterfaceState.DOWN,
                }
            ]
        }
    )
    saved_state = statelib.show_saved_config_only((VLAN_IFNAME,))
    iface_state = saved_state[Interface.KEY][0]

    assert saved_state[Interface.KEY][0][Interface.STATE] == InterfaceState.UP
    assert iface_state[VLAN.CONFIG_SUBTREE] == {
        VLAN.BASE_IFACE: "eth1",
        VLAN.ID: 101,
    }
    assertlib.assert_state_match_full(saved_state, running_state)



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.schema import VLAN
from .common import NM


def create_setting(iface_state, base_con_profile):
    vlan = iface_state.get(VLAN.TYPE)
    if not vlan:
        return None

    vlan_id = vlan[VLAN.ID]
    vlan_base_iface = vlan[VLAN.BASE_IFACE]

    vlan_setting = None
    if base_con_profile:
        vlan_setting = base_con_profile.get_setting_vlan()
        if vlan_setting:
            vlan_setting = vlan_setting.duplicate()

    if not vlan_setting:
        vlan_setting = NM.SettingVlan.new()

    vlan_setting.props.id = vlan_id
    vlan_setting.props.parent = vlan_base_iface

    return vlan_setting


def get_vlan_config(nm_profile):
    nm_setting = nm_profile.get_setting_vlan()
    if nm_setting:
        return {
            VLAN.ID: nm_setting.props.id,
            VLAN.BASE_IFACE: nm_setting.props.parent,
        }
    return {}



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

# This file will hold the common code shared by linux bridge and ovs bridge.

from operator import itemgetter

from libnmstate.schema import Bridge

from ..state import merge_dict
from .base_iface import BaseIface


class BridgeIface(BaseIface):
    BRPORT_OPTIONS_METADATA = "_brport_options"

    @property
    def is_controller(self):
        return True

    @property
    def is_virtual(self):
        return True

    def sort_port(self):
        if self.port:
            self.raw[Bridge.CONFIG_SUBTREE][Bridge.PORT_SUBTREE].sort(
                key=itemgetter(Bridge.Port.NAME)
            )

    @property
    def _bridge_config(self):
        return self.raw.get(Bridge.CONFIG_SUBTREE, {})

    @property
    def port_configs(self):
        return self._bridge_config.get(Bridge.PORT_SUBTREE, [])

    def merge(self, other):
        super().merge(other)
        self._merge_bridge_ports(other)

    def _merge_bridge_ports(self, other):
        """
        Given a bridge desired state, and it's current state, merges
        those together.

        This extension of the interface merging mechanism simplifies the user's
        life in scenarios where the user wants to partially update the bridge's
        configuration - e.g. update only the bridge's port STP configuration -
        since it enables the user to simply specify the updated values rather
        than the full current state + the updated value.
        """
        if self._bridge_config.get(Bridge.PORT_SUBTREE) == []:
            # User explictly defined empty list for ports and expecting
            # removal of all ports.
            return

        other_indexed_ports = _index_port_configs(other.port_configs)
        self_indexed_ports = _index_port_configs(self.port_configs)

        # When defined, user need to specify the whole list of ports.
        for port_iface_name in (
            other_indexed_ports.keys() & self_indexed_ports.keys()
        ):
            merge_dict(
                self_indexed_ports[port_iface_name],
                other_indexed_ports[port_iface_name],
            )
        self.raw[Bridge.CONFIG_SUBTREE][Bridge.PORT_SUBTREE] = list(
            self_indexed_ports.values()
        )

    def pre_edit_validation_and_cleanup(self):
        if self.is_up:
            self.sort_port()
        super().pre_edit_validation_and_cleanup()

    def config_changed_port(self, cur_iface):
        changed_port = []
        cur_indexed_ports = _index_port_configs(cur_iface.port_configs)
        for port_config in self.port_configs:
            port_name = port_config[Bridge.Port.NAME]
            cur_port_config = cur_indexed_ports.get(port_name)
            if cur_port_config != port_config:
                changed_port.append(port_name)
        return changed_port


def _index_port_configs(port_configs):
    return {port[Bridge.Port.NAME]: port for port in port_configs}



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.schema import VRF
from .common import NM


def create_vrf_setting(vrf_config):
    vrf_setting = NM.SettingVrf.new()
    vrf_setting.props.table = vrf_config[VRF.ROUTE_TABLE_ID]
    return vrf_setting


def get_vrf_config(nm_profile, subordinate_nm_profiles):
    nm_setting = nm_profile.get_setting_by_name(NM.SETTING_VRF_SETTING_NAME)
    if nm_setting:
        return {
            VRF.PORT_SUBTREE: [
                port_profile.get_interface_name()
                for port_profile in subordinate_nm_profiles
            ],
            VRF.ROUTE_TABLE_ID: nm_setting.props.table,
        }
    return {}



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

from libnmstate.schema import Bond
from libnmstate.schema import OVSBridge


def test_bond_deprecated_constants():
    with pytest.warns(FutureWarning) as record:
        deprecated_value = getattr(Bond, "SLAVES")

    assert len(record) == 1
    assert "SLAVES" in record[0].message.args[0]
    assert deprecated_value == "port"


def test_ovsbridge_slaves_subtree_deprecated_constants():
    with pytest.warns(FutureWarning) as record:
        deprecated_value = getattr(
            OVSBridge.Port.LinkAggregation, "SLAVES_SUBTREE"
        )

    assert len(record) == 1
    assert "SLAVES_SUBTREE" in record[0].message.args[0]
    assert deprecated_value == "port"


def test_ovsbridge_port_subtree_deprecated_constants():
    with pytest.warns(FutureWarning) as record:
        # pylint: disable=E1101
        deprecated_value = getattr(
            OVSBridge.Port.LinkAggregation.Slave, "NAME"
        )
        # pylint: enable=E1101

    assert len(record) == 1
    assert "Port" in record[0].message.args[0]
    assert deprecated_value == "name"



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager

import libnmstate
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState

from .testlib import statelib
from .testlib import cmdlib

_IPV4_EXTRA_CONFIG = "ipv4.dad-timeout"
_IPV4_EXTRA_VALUE = "0"
_IPV6_EXTRA_CONFIG = "ipv6.dhcp-hostname"
_IPV6_EXTRA_VALUE = "libnmstate.example.com"

IPV4_ADDRESS1 = "192.0.2.251"
IPV6_ADDRESS1 = "2001:db8:1::1"


def test_reapply_preserve_ip_config(eth1_up):
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: "eth1",
                    Interface.TYPE: InterfaceType.ETHERNET,
                    Interface.STATE: InterfaceState.UP,
                    Interface.IPV4: {
                        InterfaceIPv4.ADDRESS: [
                            {
                                InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS1,
                                InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
                            }
                        ],
                        InterfaceIPv4.ENABLED: True,
                    },
                    Interface.IPV6: {
                        InterfaceIPv6.ADDRESS: [
                            {
                                InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS1,
                                InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
                            }
                        ],
                        InterfaceIPv6.ENABLED: True,
                    },
                    Interface.MTU: 1500,
                }
            ]
        }
    )
    cur_state = statelib.show_only(("eth1",))
    iface_name = cur_state[Interface.KEY][0][Interface.NAME]

    for key, value in (
        (_IPV4_EXTRA_CONFIG, _IPV4_EXTRA_VALUE),
        (_IPV6_EXTRA_CONFIG, _IPV6_EXTRA_VALUE),
    ):
        with _extra_ip_config(iface_name, key, value):
            libnmstate.apply(cur_state)
            _assert_extra_ip_config(iface_name, key, value)


def _get_cur_extra_ip_config(profile_name, key):
    rc, output, _ = cmdlib.exec_cmd(
        ["nmcli", "--get-values", key, "connection", "show", profile_name]
    )
    assert rc == 0
    return output.split("\n")[0]


@contextmanager
def _extra_ip_config(iface_name, key, value):
    old_value = _get_cur_extra_ip_config(iface_name, key)
    _apply_extra_ip_config(iface_name, key, value)
    try:
        yield
    finally:
        _apply_extra_ip_config(iface_name, key, old_value)


def _apply_extra_ip_config(iface_name, key, value):
    assert (
        cmdlib.exec_cmd(
            ["nmcli", "connection", "modify", iface_name, key, value]
        )[0]
        == 0
    )


def _assert_extra_ip_config(iface_name, key, value):
    """
    Check whether extra config is touched by libnmstate.
    """
    cur_value = _get_cur_extra_ip_config(iface_name, key)
    assert cur_value == value



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from copy import deepcopy

from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import LinuxBridge as LB

from .ifacelib import gen_foo_iface_info
from .constants import PORT1_IFACE_NAME
from .constants import PORT2_IFACE_NAME

Port = LB.Port
Vlan = LB.Port.Vlan

LINUX_BRIDGE_IFACE_NAME = "linux-br0"

PORT1_PORT_CONFIG = {
    Port.NAME: PORT1_IFACE_NAME,
    Port.STP_HAIRPIN_MODE: False,
    Port.STP_PATH_COST: 100,
    Port.STP_PRIORITY: 32,
}

PORT1_VLAN_CONFIG_ACCESS = {
    Vlan.MODE: Vlan.Mode.ACCESS,
    Vlan.TAG: 305,
}

PORT2_PORT_CONFIG = {
    Port.NAME: PORT2_IFACE_NAME,
    Port.STP_HAIRPIN_MODE: False,
    Port.STP_PATH_COST: 100,
    Port.STP_PRIORITY: 32,
}

TRUNK_TAGS_IDS = [{Vlan.TrunkTags.ID: 101}, {Vlan.TrunkTags.ID: 102}]
TRUNK_TAGS_ID_RANGES = [
    {
        Vlan.TrunkTags.ID_RANGE: {
            Vlan.TrunkTags.MIN_RANGE: 400,
            Vlan.TrunkTags.MAX_RANGE: 500,
        }
    },
    {
        Vlan.TrunkTags.ID_RANGE: {
            Vlan.TrunkTags.MIN_RANGE: 600,
            Vlan.TrunkTags.MAX_RANGE: 900,
        }
    },
]

PORT2_VLAN_CONFIG_TRUNK_WITH_ID_RANGE = {
    Vlan.MODE: Vlan.Mode.TRUNK,
    Vlan.TAG: 105,
    Vlan.ENABLE_NATIVE: True,
    Vlan.TRUNK_TAGS: TRUNK_TAGS_ID_RANGES,
}

PORT2_VLAN_CONFIG_TRUNK_WITH_ID = {
    Vlan.MODE: Vlan.Mode.TRUNK,
    Vlan.TAG: 105,
    Vlan.ENABLE_NATIVE: True,
    Vlan.TRUNK_TAGS: TRUNK_TAGS_IDS,
}

PORT2_VLAN_CONFIG_TRUNK = PORT2_VLAN_CONFIG_TRUNK_WITH_ID

TEST_PORT_PORT_CONFIGS = [PORT1_PORT_CONFIG, PORT2_PORT_CONFIG]

TEST_PORT_NAMES = [PORT1_IFACE_NAME, PORT2_IFACE_NAME]


def gen_bridge_iface_info():
    iface_info = gen_foo_iface_info(iface_type=InterfaceType.LINUX_BRIDGE)
    iface_info[Interface.NAME] = LINUX_BRIDGE_IFACE_NAME
    iface_info[LB.CONFIG_SUBTREE] = {
        LB.PORT_SUBTREE: deepcopy(TEST_PORT_PORT_CONFIGS),
        LB.OPTIONS_SUBTREE: {
            LB.Options.GROUP_FORWARD_MASK: 0,
            LB.Options.MAC_AGEING_TIME: 300,
            LB.Options.MULTICAST_SNOOPING: True,
            LB.STP_SUBTREE: {
                LB.STP.ENABLED: False,
                LB.STP.FORWARD_DELAY: 15,
                LB.STP.HELLO_TIME: 2,
                LB.STP.MAX_AGE: 20,
                LB.STP.PRIORITY: 32768,
            },
        },
    }
    return iface_info


def gen_bridge_iface_info_with_vlan_filter():
    iface_info = gen_bridge_iface_info()
    br_config = iface_info[LB.CONFIG_SUBTREE]
    br_config[LB.PORT_SUBTREE][0][Port.VLAN_SUBTREE] = deepcopy(
        PORT1_VLAN_CONFIG_ACCESS
    )
    br_config[LB.PORT_SUBTREE][1][Port.VLAN_SUBTREE] = deepcopy(
        PORT2_VLAN_CONFIG_TRUNK
    )
    return iface_info



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from abc import ABCMeta
from abc import abstractmethod
from collections.abc import Sequence
from collections.abc import Mapping
from functools import total_ordering


@total_ordering
class StateEntry(metaclass=ABCMeta):
    @abstractmethod
    def _keys(self):
        """
        Return the tuple representing this entry, will be used for hashing or
        comparing.
        """
        pass

    def __hash__(self):
        return hash(self._keys())

    def __eq__(self, other):
        return self is other or self._keys() == other._keys()

    def __lt__(self, other):
        return self._keys() < other._keys()

    def __repr__(self):
        return str(self.to_dict())

    @property
    @abstractmethod
    def absent(self):
        pass

    def to_dict(self):
        return {
            key.replace("_", "-"): value
            for key, value in vars(self).items()
            if (not key.startswith("_")) and (value is not None)
        }

    def match(self, other):
        """
        Match self against other. Treat self None attributes as wildcards,
        matching against any value in others.
        Return True for a match, False otherwise.
        """
        for self_value, other_value in zip(self._keys(), other._keys()):
            if self_value is not None and self_value != other_value:
                return False
        return True


def state_match(desire, current):
    """
    Return True when all values defined in desire equal to value in current,
    else False:
        * For mapping(e.g. dict), desire could have less value than current.
        * For sequnce(e.g. list), desire should equal to current.
    """
    if isinstance(desire, Mapping):
        return isinstance(current, Mapping) and all(
            state_match(val, current.get(key)) for key, val in desire.items()
        )
    elif isinstance(desire, Sequence) and not isinstance(desire, str):
        return (
            isinstance(current, Sequence)
            and not isinstance(current, str)
            and len(current) == len(desire)
            and all(state_match(d, c) for d, c in zip(desire, current))
        )
    else:
        return desire == current


def merge_dict(dict_to, dict_from):
    """
    Data will copy from `dict_from` if undefined in `dict_to`.
    For list, the whole list is copied instead of merging.
    """
    for key, from_value in dict_from.items():
        if key not in dict_to:
            dict_to[key] = from_value
        elif isinstance(dict_to[key], Mapping):
            merge_dict(dict_to[key], from_value)



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.schema import Ethernet

from .common import NM
from .common import GLib


SRIOV_NMSTATE_TO_NM_MAP = {
    Ethernet.SRIOV.VFS.MAC_ADDRESS: (
        NM.SRIOV_VF_ATTRIBUTE_MAC,
        GLib.Variant.new_string,
    ),
    Ethernet.SRIOV.VFS.SPOOF_CHECK: (
        NM.SRIOV_VF_ATTRIBUTE_SPOOF_CHECK,
        GLib.Variant.new_boolean,
    ),
    Ethernet.SRIOV.VFS.TRUST: (
        NM.SRIOV_VF_ATTRIBUTE_TRUST,
        GLib.Variant.new_boolean,
    ),
    Ethernet.SRIOV.VFS.MIN_TX_RATE: (
        NM.SRIOV_VF_ATTRIBUTE_MIN_TX_RATE,
        GLib.Variant.new_uint32,
    ),
    Ethernet.SRIOV.VFS.MAX_TX_RATE: (
        NM.SRIOV_VF_ATTRIBUTE_MAX_TX_RATE,
        GLib.Variant.new_uint32,
    ),
}


def create_setting(iface_state, base_con_profile):
    sriov_setting = None
    sriov_config = iface_state.get(Ethernet.CONFIG_SUBTREE, {}).get(
        Ethernet.SRIOV_SUBTREE
    )

    if base_con_profile:
        sriov_setting = base_con_profile.get_setting_by_name(
            NM.SETTING_SRIOV_SETTING_NAME
        )
    if sriov_config:
        if sriov_setting:
            sriov_setting = sriov_setting.duplicate()
        else:
            sriov_setting = NM.SettingSriov.new()

        vfs_config = sriov_config.get(Ethernet.SRIOV.VFS_SUBTREE, [])
        vf_object_ids = {vf.get_index() for vf in sriov_setting.props.vfs}
        vf_config_ids = {
            vf_config[Ethernet.SRIOV.VFS.ID] for vf_config in vfs_config
        }

        # As the user must do full edit of vfs, nmstate is deleting all the vfs
        # and then adding all the vfs from the config.
        for vf_id in _remove_sriov_vfs_in_setting(
            vfs_config, sriov_setting, vf_object_ids
        ):
            sriov_setting.remove_vf_by_index(vf_id)

        for vf_object in _create_sriov_vfs_from_config(
            vfs_config, sriov_setting, vf_config_ids
        ):
            sriov_setting.add_vf(vf_object)

        sriov_setting.props.total_vfs = sriov_config[Ethernet.SRIOV.TOTAL_VFS]

    return sriov_setting


def _create_sriov_vfs_from_config(vfs_config, sriov_setting, vf_ids_to_add):
    vfs_config_to_add = (
        vf_config
        for vf_config in vfs_config
        if vf_config[Ethernet.SRIOV.VFS.ID] in vf_ids_to_add
    )
    for vf_config in vfs_config_to_add:
        vf_id = vf_config.pop(Ethernet.SRIOV.VFS.ID)
        vf_object = NM.SriovVF.new(vf_id)
        for key, val in vf_config.items():
            _set_nm_attribute(vf_object, key, val)

        yield vf_object


def _set_nm_attribute(vf_object, key, value):
    nm_attr, nm_variant = SRIOV_NMSTATE_TO_NM_MAP[key]
    vf_object.set_attribute(nm_attr, nm_variant(value))


def _remove_sriov_vfs_in_setting(vfs_config, sriov_setting, vf_ids_to_remove):
    for vf_id in vf_ids_to_remove:
        yield vf_id


def get_sriov_config(nm_profile):
    nm_setting = nm_profile.get_setting_by_name(NM.SETTING_SRIOV_SETTING_NAME)
    if nm_setting:
        vfs_info = []
        for nm_vf in nm_setting.props.vfs:
            vf_info = {Ethernet.SRIOV.VFS.ID: nm_vf.get_index()}
            attributes = nm_vf.get_attribute_names()
            for key in (
                Ethernet.SRIOV.VFS.SPOOF_CHECK,
                Ethernet.SRIOV.VFS.MAC_ADDRESS,
                Ethernet.SRIOV.VFS.TRUST,
            ):
                if key in attributes:
                    vf_info[key] = nm_vf.get_attribute(key)

            vfs_info.append(vf_info)

        return {
            Ethernet.SRIOV.TOTAL_VFS: nm_setting.props.total_vfs,
            Ethernet.SRIOV.VFS_SUBTREE: vfs_info,
        }
    return {}



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from copy import deepcopy

from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import OVSBridge

from .ifacelib import gen_foo_iface_info
from .constants import PORT1_IFACE_NAME
from .constants import PORT2_IFACE_NAME

OVS_BRIDGE_IFACE_NAME = "ovs-br0"
OVS_IFACE_NAME = "ovs0"

PORT_PORT_CONFIGS = [
    {OVSBridge.Port.NAME: PORT1_IFACE_NAME},
    {OVSBridge.Port.NAME: PORT2_IFACE_NAME},
    {OVSBridge.Port.NAME: OVS_IFACE_NAME},
]


def gen_ovs_bridge_info():
    iface_info = gen_foo_iface_info(iface_type=InterfaceType.OVS_BRIDGE)
    iface_info[Interface.NAME] = OVS_BRIDGE_IFACE_NAME
    iface_info[OVSBridge.CONFIG_SUBTREE] = {
        OVSBridge.PORT_SUBTREE: deepcopy(PORT_PORT_CONFIGS),
        OVSBridge.OPTIONS_SUBTREE: {},
    }
    return iface_info



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager
from subprocess import CalledProcessError

from . import cmdlib


@contextmanager
def disable_nm_plugin(plugin):
    try:
        _, rpm_output, _ = cmdlib.exec_cmd(
            ("rpm", "-ql", f"NetworkManager-{plugin}"), check=True
        )
    except CalledProcessError:
        yield
    else:
        lib_path = [lib for lib in rpm_output.split() if lib.endswith(".so")][
            0
        ]
        with mount_devnull_to_path(lib_path):
            yield


@contextmanager
def mount_devnull_to_path(lib_path):
    try:
        cmdlib.exec_cmd(("mount", "--bind", "/dev/null", lib_path), check=True)
        with nm_service_restart():
            try:
                yield
            finally:
                cmdlib.exec_cmd(("umount", lib_path), check=True)
    except CalledProcessError:
        cmdlib.exec_cmd(("umount", lib_path), check=True)
        raise


@contextmanager
def nm_service_restart():
    systemctl_restart_nm_cmd = ("systemctl", "restart", "NetworkManager")
    cmdlib.exec_cmd(systemctl_restart_nm_cmd, check=True)
    try:
        yield
    finally:
        cmdlib.exec_cmd(systemctl_restart_nm_cmd, check=True)



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.error import NmstateValueError
from libnmstate.schema import InfiniBand
from libnmstate.schema import Interface

from .base_iface import BaseIface


class InfiniBandIface(BaseIface):
    @property
    def parent(self):
        return self._ib_config.get(InfiniBand.BASE_IFACE)

    @property
    def _pkey(self):
        return self._ib_config.get(InfiniBand.PKEY)

    @property
    def need_parent(self):
        if not self._pkey or str(self._pkey) == str(InfiniBand.DEFAULT_PKEY):
            return False
        return True

    @property
    def _ib_config(self):
        return self.raw.get(InfiniBand.CONFIG_SUBTREE, {})

    def pre_edit_validation_and_cleanup(self):
        if self.is_up:
            _cannonicalize_pkey(self.raw)
            self._validate_mandatory_properties()
            self._validate_interface_name_format_for_pkey_nic()
        super().pre_edit_validation_and_cleanup()

    def _validate_mandatory_properties(self):
        if self.is_up:
            if InfiniBand.MODE not in self._ib_config:
                raise NmstateValueError(
                    f"InfiniBand interface {self.name} has missing "
                    f"mandatory property: {InfiniBand.MODE}"
                )

    def state_for_verify(self):
        state = super().state_for_verify()
        _cannonicalize_pkey(state)
        return state

    def _validate_interface_name_format_for_pkey_nic(self):
        if self.need_parent:
            pkey_str = f"{self._pkey:x}"
            expected_name = f"{self.parent}.{pkey_str}"
            if self.name != expected_name:
                raise NmstateValueError(
                    f"InfiniBand interface name {self.name} is invalid "
                    f"for specified pkey, should be {expected_name}"
                )


def _cannonicalize_pkey(iface_info):
    """
    * Set as 0xffff if pkey not defined
    * Convert pkey string to integer
    * Raise NmstateValueError when out of range(16 bites)
    """
    iface_name = iface_info[Interface.NAME]
    ib_config = iface_info.get(InfiniBand.CONFIG_SUBTREE, {})
    original_pkey = ib_config.get(InfiniBand.PKEY)
    is_invalid = False
    if original_pkey is None:
        ib_config[InfiniBand.PKEY] = 0xFFFF
    elif isinstance(original_pkey, str):
        if original_pkey.startswith("0x"):
            try:
                ib_config[InfiniBand.PKEY] = int(original_pkey, 16)
            except ValueError:
                is_invalid = True
        else:
            try:
                ib_config[InfiniBand.PKEY] = int(original_pkey, 10)
            except ValueError:
                is_invalid = True

    if (
        is_invalid
        or ib_config[InfiniBand.PKEY] > 0xFFFF
        or ib_config[InfiniBand.PKEY] <= 0
    ):
        raise NmstateValueError(
            f"Invalid infiniband pkey for interface {iface_name}: "
            f"{original_pkey}, should be integer or hex string in the range of"
            "1 - 0xffff"
        )



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

from libnmstate.error import NmstateValueError
from libnmstate.schema import VRF
from libnmstate.schema import InterfaceType

from libnmstate.ifaces.vrf import VrfIface

from ..testlib.ifacelib import gen_foo_iface_info

PORT1_IFACE_NAME = "port1"
PORT2_IFACE_NAME = "port2"


class TestVrfIface:
    def _gen_iface_info(self):
        iface_info = gen_foo_iface_info(iface_type=InterfaceType.TEAM)
        iface_info[VRF.CONFIG_SUBTREE] = {
            VRF.PORT_SUBTREE: [PORT1_IFACE_NAME, PORT2_IFACE_NAME],
            VRF.ROUTE_TABLE_ID: 100,
        }
        return iface_info

    def test_vrf_is_virtual(self):
        assert VrfIface(self._gen_iface_info()).is_virtual

    def test_vrf_is_controller(self):
        assert VrfIface(self._gen_iface_info()).is_controller

    def test_vrf_sort_ports(self):
        iface1_info = self._gen_iface_info()
        iface2_info = self._gen_iface_info()
        iface2_info[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE].reverse()

        iface1 = VrfIface(iface1_info)
        iface2 = VrfIface(iface2_info)

        assert iface1.state_for_verify() == iface2.state_for_verify()

    def test_vrf_remove_port(self):
        iface_info = self._gen_iface_info()
        iface_info[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE].pop()

        iface = VrfIface(iface_info)
        assert iface.port == [PORT1_IFACE_NAME]

    def test_validate_missing_table_id(self):
        iface_info = self._gen_iface_info()
        iface_info[VRF.CONFIG_SUBTREE].pop(VRF.ROUTE_TABLE_ID)
        iface = VrfIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager

import libnmstate
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState

from . import cmdlib


@contextmanager
def nm_unmanaged_dummy(name):
    cmdlib.exec_cmd(f"ip link add name {name} type dummy".split(), check=True)
    cmdlib.exec_cmd(f"ip link set {name} up".split(), check=True)
    cmdlib.exec_cmd(f"nmcli d set {name} managed false".split(), check=True)
    try:
        yield
    finally:
        try:
            libnmstate.apply(
                {
                    Interface.KEY: [
                        {
                            Interface.NAME: name,
                            Interface.STATE: InterfaceState.ABSENT,
                        }
                    ]
                },
                verify_change=False,
            )
        except Exception:
            # dummy1 might not became managed by NM, hence removal might fail
            cmdlib.exec_cmd(f"ip link del {name}".split())



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.schema import Ethernet
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState

from .base_iface import BaseIface


class EthernetIface(BaseIface):
    def merge(self, other):
        """
        Given the other_state, update the ethernet interfaces state base on
        the other_state ethernet interfaces data.
        Usually the other_state represents the current state.
        If auto-negotiation, speed and duplex settings are not provided,
        but exist in the current state, they need to be set to None
        to not override them with the values from the current settings
        since the current settings are read from the device state and not
        from the actual configuration.  This makes it possible to distinguish
        whether a user specified these values in the later configuration step.
        """
        eth_conf = self._info.setdefault(Ethernet.CONFIG_SUBTREE, {})
        eth_conf.setdefault(Ethernet.AUTO_NEGOTIATION, None)
        eth_conf.setdefault(Ethernet.SPEED, None)
        eth_conf.setdefault(Ethernet.DUPLEX, None)
        super().merge(other)

    def state_for_verify(self):
        state = super().state_for_verify()
        _capitalize_sriov_vf_mac(state)
        return state

    @property
    def sriov_total_vfs(self):
        return (
            self.raw.get(Ethernet.CONFIG_SUBTREE, {})
            .get(Ethernet.SRIOV_SUBTREE, {})
            .get(Ethernet.SRIOV.TOTAL_VFS, 0)
        )

    def create_sriov_vf_ifaces(self):
        return [
            EthernetIface(
                {
                    # According to manpage of systemd.net-naming-scheme(7),
                    # SRIOV VF interface will have v{slot} in device name.
                    # Currently, nmstate has no intention to support
                    # user-defined udev rule on SRIOV interface naming policy.
                    Interface.NAME: f"{self.name}v{i}",
                    Interface.TYPE: InterfaceType.ETHERNET,
                    # VF will be in DOWN state initialy.
                    Interface.STATE: InterfaceState.DOWN,
                }
            )
            for i in range(0, self.sriov_total_vfs)
        ]


def _capitalize_sriov_vf_mac(state):
    vfs = (
        state.get(Ethernet.CONFIG_SUBTREE, {})
        .get(Ethernet.SRIOV_SUBTREE, {})
        .get(Ethernet.SRIOV.VFS_SUBTREE, [])
    )
    for vf in vfs:
        vf_mac = vf.get(Ethernet.SRIOV.VFS.MAC_ADDRESS)
        if vf_mac:
            vf[Ethernet.SRIOV.VFS.MAC_ADDRESS] = vf_mac.upper()



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
import copy
import time

import pytest
import jsonschema as js

import libnmstate
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState
from libnmstate.error import NmstateVerificationError

from .testlib import assertlib
from .testlib import statelib
from .testlib.iproutelib import ip_monitor_assert_stable_link_up
from .testlib.vlan import vlan_interface


@pytest.fixture(scope="function", autouse=True)
def eth1(eth1_up):
    pass


@pytest.fixture
def eth1_with_ipv6(eth1_up):
    ifstate = eth1_up[Interface.KEY][0]
    ifstate[Interface.IPV6][InterfaceIPv6.ENABLED] = True
    libnmstate.apply(eth1_up)


@pytest.mark.tier1
def test_increase_iface_mtu():
    desired_state = statelib.show_only(("eth1",))
    eth1_desired_state = desired_state[Interface.KEY][0]
    eth1_desired_state[Interface.MTU] = 1900

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


@pytest.mark.tier1
def test_decrease_iface_mtu():
    desired_state = statelib.show_only(("eth1",))
    eth1_desired_state = desired_state[Interface.KEY][0]
    eth1_desired_state[Interface.MTU] = 1400

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


@pytest.mark.tier1
def test_upper_limit_jambo_iface_mtu():
    desired_state = statelib.show_only(("eth1",))
    eth1_desired_state = desired_state[Interface.KEY][0]
    eth1_desired_state[Interface.MTU] = 9000

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


def test_increase_more_than_jambo_iface_mtu():
    desired_state = statelib.show_only(("eth1",))
    eth1_desired_state = desired_state[Interface.KEY][0]
    eth1_desired_state[Interface.MTU] = 10000

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


def test_decrease_to_zero_iface_mtu():
    desired_state = statelib.show_only(("eth1",))
    origin_desired_state = copy.deepcopy(desired_state)
    eth1_desired_state = desired_state[Interface.KEY][0]
    eth1_desired_state[Interface.MTU] = 0

    with pytest.raises(NmstateVerificationError) as err:
        libnmstate.apply(desired_state)
    assert "-mtu: 0" in err.value.args[0]
    # FIXME: Drop the sleep when the waiting logic is implemented.
    time.sleep(2)
    assertlib.assert_state(origin_desired_state)


def test_decrease_to_negative_iface_mtu():
    desired_state = statelib.show_only(("eth1",))
    origin_desired_state = copy.deepcopy(desired_state)
    eth1_desired_state = desired_state[Interface.KEY][0]
    eth1_desired_state[Interface.MTU] = -1

    with pytest.raises(js.ValidationError) as err:
        libnmstate.apply(desired_state)
    assert "-1" in err.value.args[0]
    assertlib.assert_state(origin_desired_state)


def test_decrease_to_ipv6_min_ethernet_frame_size_iface_mtu(eth1_with_ipv6):
    desired_state = statelib.show_only(("eth1",))
    eth1_desired_state = desired_state[Interface.KEY][0]
    eth1_desired_state[Interface.MTU] = 1280

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


def test_decrease_to_lower_than_min_ipv6_iface_mtu(eth1_with_ipv6):
    original_state = statelib.show_only(("eth1",))
    desired_state = copy.deepcopy(original_state)
    eth1_desired_state = desired_state[Interface.KEY][0]
    eth1_desired_state[Interface.MTU] = 1279

    with pytest.raises(NmstateVerificationError) as err:
        libnmstate.apply(desired_state)
    assert "1279" in err.value.args[0]
    # FIXME: Drop the sleep when the waiting logic is implemented.
    time.sleep(2)
    assertlib.assert_state(original_state)


def test_mtu_without_ipv6(eth1_up):
    eth1_up[Interface.KEY][0][Interface.MTU] = 576
    libnmstate.apply(eth1_up)
    assertlib.assert_state(eth1_up)


@pytest.mark.tier1
def test_set_mtu_on_two_vlans_with_a_shared_base(eth1_up):
    base_ifname = eth1_up[Interface.KEY][0][Interface.NAME]
    v101 = vlan_interface("eth1.101", 101, base_ifname)
    v102 = vlan_interface("eth1.102", 102, base_ifname)
    with v101 as v101_state, v102 as v102_state:
        desired_state = {
            Interface.KEY: [
                eth1_up[Interface.KEY][0],
                v101_state[Interface.KEY][0],
                v102_state[Interface.KEY][0],
            ]
        }
        for iface_state in desired_state[Interface.KEY]:
            iface_state[Interface.MTU] = 2000

        libnmstate.apply(desired_state)

        assertlib.assert_state(desired_state)


@pytest.mark.tier1
@ip_monitor_assert_stable_link_up("eth1")
def test_change_mtu_with_stable_link_up(eth1_up):
    desired_state = statelib.show_only(("eth1",))
    eth1_desired_state = desired_state[Interface.KEY][0]
    eth1_desired_state[Interface.MTU] = 1900

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


@pytest.fixture(scope="function")
def eth1_up_with_mtu_1900(eth1_up):
    desired_state = statelib.show_only(("eth1",))
    eth1_desired_state = desired_state[Interface.KEY][0]
    eth1_desired_state[Interface.MTU] = 1900

    libnmstate.apply(desired_state)
    yield desired_state


def test_empty_state_preserve_the_old_mtu(eth1_up_with_mtu_1900):
    desired_state = eth1_up_with_mtu_1900
    libnmstate.apply({Interface.KEY: [{Interface.NAME: "eth1"}]})

    assertlib.assert_state(desired_state)


def test_show_saved_config_on_ethernet_with_mtu(eth1_up_with_mtu_1900):
    running_state = statelib.show_only(("eth1",))
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: "eth1",
                    Interface.STATE: InterfaceState.DOWN,
                }
            ]
        }
    )
    saved_state = statelib.show_saved_config_only(("eth1",))

    assert saved_state[Interface.KEY][0][Interface.MTU] == 1900
    assert saved_state[Interface.KEY][0][Interface.STATE] == InterfaceState.UP
    assertlib.assert_state_match_full(saved_state, running_state)



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

import libnmstate
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6

from ..testlib import cmdlib
from ..testlib import assertlib


def test_get_applied_config_for_dhcp_state_with_dhcp_enabeld_on_disk(eth1_up):
    iface_state = eth1_up[Interface.KEY][0]
    iface_name = iface_state[Interface.NAME]
    cmdlib.exec_cmd(
        f"nmcli c modify {iface_name} ipv4.method auto".split(), check=True
    )
    cmdlib.exec_cmd(
        f"nmcli c modify {iface_name} ipv6.method auto".split(), check=True
    )

    assertlib.assert_state_match({Interface.KEY: [iface_state]})


@pytest.fixture
def eth1_up_with_auto_ip(eth1_up):
    iface_name = eth1_up[Interface.KEY][0][Interface.NAME]
    iface_state = {
        Interface.NAME: iface_name,
        Interface.IPV4: {
            InterfaceIPv4.ENABLED: True,
            InterfaceIPv4.DHCP: True,
        },
        Interface.IPV6: {
            InterfaceIPv6.ENABLED: True,
            InterfaceIPv6.DHCP: True,
            InterfaceIPv6.AUTOCONF: True,
        },
    }
    libnmstate.apply({Interface.KEY: [iface_state]})
    yield iface_state


def test_get_applied_config_for_dhcp_state_with_dhcp_disabled_on_disk(
    eth1_up_with_auto_ip,
):
    iface_state = eth1_up_with_auto_ip
    iface_name = iface_state[Interface.NAME]
    cmdlib.exec_cmd(
        f"nmcli c modify {iface_name} ipv4.method disabled".split(), check=True
    )
    cmdlib.exec_cmd(
        f"nmcli c modify {iface_name} ipv6.method disabled".split(), check=True
    )

    assertlib.assert_state_match({Interface.KEY: [iface_state]})



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

# This file is targeting:
#   * NM.RemoteConnection, NM.SimpleConnection releated

from distutils.version import StrictVersion

from libnmstate.error import NmstateLibnmError
from libnmstate.error import NmstateNotSupportedError
from libnmstate.error import NmstateInternalError
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState
from libnmstate.ifaces.base_iface import BaseIface

from .active_connection import ActiveConnectionDeactivate
from .active_connection import ProfileActivation
from .active_connection import is_activated
from .common import NM
from .connection import create_new_nm_simple_conn
from .device import get_nm_dev
from .device import DeviceReapply
from .device import DeviceDelete
from .translator import Api2Nm
from .macvlan import is_macvtap


class NmProfile:
    # For unmanged iface and desired to down
    ACTION_ACTIVATE_FIRST = "activate_first"
    ACTION_DEACTIVATE = "deactivate"
    ACTION_DEACTIVATE_FIRST = "deactivate_first"
    ACTION_DELETE_DEVICE = "delete_device"
    ACTION_MODIFIED = "modified"
    ACTION_NEW_IFACES = "new_ifaces"
    ACTION_NEW_OVS_IFACE = "new_ovs_iface"
    ACTION_NEW_OVS_PORT = "new_ovs_port"
    ACTION_NEW_VLAN = "new_vlan"
    ACTION_NEW_VXLAN = "new_vxlan"
    ACTION_OTHER_MASTER = "other_master"
    ACTION_DELETE_PROFILE = "delete_profile"
    ACTION_TOP_MASTER = "top_master"

    # This is order on group for activation/deactivation
    ACTIONS = (
        ACTION_ACTIVATE_FIRST,
        ACTION_DEACTIVATE_FIRST,
        ACTION_TOP_MASTER,
        ACTION_NEW_IFACES,
        ACTION_OTHER_MASTER,
        ACTION_NEW_OVS_PORT,
        ACTION_NEW_OVS_IFACE,
        ACTION_MODIFIED,
        ACTION_NEW_VLAN,
        ACTION_NEW_VXLAN,
        ACTION_DEACTIVATE,
        ACTION_DELETE_PROFILE,
        ACTION_DELETE_DEVICE,
    )

    def __init__(self, ctx, iface, save_to_disk):
        self._ctx = ctx
        self._iface = iface
        self._save_to_disk = save_to_disk
        self._nm_iface_type = None
        if self._iface.type != InterfaceType.UNKNOWN:
            self._nm_iface_type = Api2Nm.get_iface_type(self._iface.type)
        self._nm_ac = None
        self._nm_dev = None
        self._nm_profile = None
        self._nm_simple_conn = None
        self._actions = set()
        self._activated = False
        self._deactivated = False
        self._profile_deleted = False
        self._device_deleted = False
        self._import_current()
        self._gen_actions()

    def _gen_actions(self):
        if self._iface.is_absent:
            self._add_action(NmProfile.ACTION_DELETE_PROFILE)
            if self._iface.is_virtual and self._nm_dev:
                self._add_action(NmProfile.ACTION_DELETE_DEVICE)
        elif self._iface.is_up:
            self._add_action(NmProfile.ACTION_MODIFIED)
            if not self._nm_dev:
                if self._iface.type == InterfaceType.OVS_PORT:
                    self._add_action(NmProfile.ACTION_NEW_OVS_PORT)
                elif self._iface.type == InterfaceType.OVS_INTERFACE:
                    self._add_action(NmProfile.ACTION_NEW_OVS_IFACE)
                elif self._iface.type == InterfaceType.VLAN:
                    self._add_action(NmProfile.ACTION_NEW_VLAN)
                elif self._iface.type == InterfaceType.VXLAN:
                    self._add_action(NmProfile.ACTION_NEW_VXLAN)
                else:
                    self._add_action(NmProfile.ACTION_NEW_IFACES)

        elif self._iface.is_down:
            if self._nm_ac:
                self._add_action(NmProfile.ACTION_DEACTIVATE)
            elif self._iface.is_virtual and self._nm_dev:
                self._add_action(NmProfile.ACTION_DELETE_DEVICE)

        if self._iface.is_controller and self._iface.is_up:
            if self._iface.controller:
                self._add_action(NmProfile.ACTION_OTHER_MASTER)
            else:
                self._add_action(NmProfile.ACTION_TOP_MASTER)

        if (
            self._iface.is_up
            and self._iface.type == InterfaceType.BOND
            and self._iface.is_bond_mode_changed
        ):
            # NetworkManager leaves leftover in sysfs for bond
            # options when changing bond mode, bug:
            # https://bugzilla.redhat.com/show_bug.cgi?id=1819137
            # Workaround: delete the bond interface from kernel and
            # create again via full deactivation beforehand.
            self._add_action(NmProfile.ACTION_DEACTIVATE_FIRST)

        if self._iface.is_up and self._iface.type in (
            InterfaceType.MAC_VLAN,
            InterfaceType.MAC_VTAP,
        ):
            # NetworkManager requires the profile to be deactivated in
            # order to modify it. Therefore if the profile is modified
            # it needs to be deactivated beforehand in order to apply
            # the changes and activate it again.
            self._add_action(NmProfile.ACTION_DEACTIVATE_FIRST)

        if (
            self._iface.is_down
            and self._nm_dev
            and not self._nm_dev.get_managed()
        ):
            # In order to deactivate an unmanaged interface, we have to
            # activate the newly created profile to remove all kernel
            # settings.
            self._add_action(NmProfile.ACTION_ACTIVATE_FIRST)

    def save_config(self):
        if self._iface.is_absent or self._iface.is_down:
            return

        self._import_current()
        self._check_sriov_support()
        self._check_unsupported_memory_only()
        # Don't create new profile if original desire does not ask
        # anything besides state:up and not been marked as changed.
        # We don't need to do this once we support querying on-disk
        # configure
        if (
            self._nm_profile is None
            and not self._iface.is_changed
            and set(self._iface.original_dict)
            <= set([Interface.STATE, Interface.NAME, Interface.TYPE])
        ):
            cur_nm_profile = self._get_first_nm_profile()
            if (
                cur_nm_profile
                and _is_memory_only(cur_nm_profile) != self._save_to_disk
            ):
                self._nm_profile = cur_nm_profile
                return

        # TODO: Use applied config as base profile
        #       Or even better remove the base profile argument as top level
        #       of nmstate should provide full/merged configure.
        self._nm_simple_conn = create_new_nm_simple_conn(
            self._iface, self._nm_profile
        )
        if self._nm_profile:
            ProfileUpdate(
                self._ctx,
                self._iface.name,
                self._iface.type,
                self._nm_simple_conn,
                self._nm_profile,
                self._save_to_disk,
            ).run()
        else:
            ProfileAdd(
                self._ctx,
                self._iface.name,
                self._iface.type,
                self._nm_simple_conn,
                self._save_to_disk,
            ).run()

    def _check_unsupported_memory_only(self):
        if (
            not self._save_to_disk
            and StrictVersion(self._ctx.client.get_version())
            < StrictVersion("1.28.0")
            and self._iface.type
            in (
                InterfaceType.OVS_BRIDGE,
                InterfaceType.OVS_INTERFACE,
                InterfaceType.OVS_PORT,
            )
        ):
            raise NmstateNotSupportedError(
                f"NetworkManager version {self._ctx.client.get_version()}"
                f" does not support 'save_to_disk=False' against"
                " OpenvSwitch interface."
            )

    def _check_sriov_support(self):
        if (
            self._nm_dev
            and self._iface.type == InterfaceType.ETHERNET
            and self._iface.sriov_total_vfs
        ):
            if (
                not self._nm_dev.props.capabilities
                & NM.DeviceCapabilities.SRIOV
            ):
                raise NmstateNotSupportedError(
                    f"Interface {self._iface.name}  {self._iface.type} "
                    "does not support SR-IOV"
                )

    def _activate(self):
        if self._activated:
            return

        if not self._nm_profile:
            self._import_nm_profile_by_simple_conn()

        profile_activation = ProfileActivation(
            self._ctx,
            self._iface.name,
            self._iface.type,
            self._nm_profile,
            self._nm_dev,
        )
        if is_activated(self._nm_ac, self._nm_dev):
            # After ProfileUpdate(), the self._nm_profile is still hold
            # the old settings, DeviceReapply should use the
            # self._nm_simple_conn for updated settings.
            DeviceReapply(
                self._ctx,
                self._iface.name,
                self._iface.type,
                self._nm_dev,
                self._nm_simple_conn,
                profile_activation,
            ).run()
        else:
            profile_activation.run()
        self._activated = True

    def _deactivate(self):
        if self._deactivated:
            return
        self._import_current()
        if self._nm_ac:
            ActiveConnectionDeactivate(
                self._ctx, self._iface.name, self._iface.type, self._nm_ac
            ).run()
        self._deactivated = True

    def _delete_profile(self):
        if self._profile_deleted:
            return
        self._import_current()
        if self._nm_profile:
            ProfileDelete(
                self._ctx, self._iface.name, self._iface.type, self._nm_profile
            ).run()

        self._profile_deleted = True

    def _delete_device(self):
        if self._device_deleted:
            return
        self._import_current()
        if self._nm_dev:
            DeviceDelete(
                self._ctx, self._iface.name, self._iface.type, self._nm_dev
            ).run()
        self._device_deleted = True

    def _add_action(self, action):
        self._actions.add(action)

    def has_action(self, action):
        return action in self._actions

    def do_action(self, action):
        if action in (
            NmProfile.ACTION_MODIFIED,
            NmProfile.ACTION_ACTIVATE_FIRST,
            NmProfile.ACTION_TOP_MASTER,
            NmProfile.ACTION_NEW_IFACES,
            NmProfile.ACTION_OTHER_MASTER,
            NmProfile.ACTION_NEW_OVS_PORT,
            NmProfile.ACTION_NEW_OVS_IFACE,
            NmProfile.ACTION_NEW_VLAN,
            NmProfile.ACTION_NEW_VXLAN,
        ):
            self._activate()
        elif (
            action
            in (
                NmProfile.ACTION_DELETE_PROFILE,
                NmProfile.ACTION_DELETE_DEVICE,
                NmProfile.ACTION_DEACTIVATE,
                NmProfile.ACTION_DEACTIVATE_FIRST,
            )
            and not self._deactivated
        ):
            self._deactivate()
        elif action == NmProfile.ACTION_DELETE_PROFILE:
            self._delete_profile()
        elif action == NmProfile.ACTION_DELETE_DEVICE:
            self._delete_device()
        else:
            raise NmstateInternalError(
                f"BUG: NmProfile.do_action() got unknown action {action}"
            )

    def _import_current(self):
        self._nm_dev = get_nm_dev(
            self._ctx, self._iface.name, self._iface.type
        )
        self._nm_ac = (
            self._nm_dev.get_active_connection() if self._nm_dev else None
        )
        self._nm_profile = (
            self._nm_ac.get_connection() if self._nm_ac else None
        )

    def _import_nm_profile_by_simple_conn(self):
        self._ctx.refresh_content()
        for nm_profile in self._ctx.client.get_connections():
            if nm_profile.get_uuid() == self._nm_simple_conn.get_uuid():
                self._nm_profile = nm_profile

    def _get_first_nm_profile(self):
        for nm_profile in self._ctx.client.get_connections():
            if nm_profile.get_interface_name() == self._iface.name and (
                self._nm_iface_type is None
                or nm_profile.get_connection_type() == self._nm_iface_type
            ):
                return nm_profile
        return None

    def delete_other_profiles(self):
        """
        Remove all profiles except the NM.RemoteConnection used by current
        NM.ActiveConnection if interface is marked as UP
        """
        if self._iface.is_down:
            return
        self._import_current()
        for nm_profile in self._ctx.client.get_connections():
            if (
                nm_profile.get_interface_name() == self._iface.name
                and (
                    self._nm_iface_type is None
                    or nm_profile.get_connection_type() == self._nm_iface_type
                )
                and (
                    self._nm_profile is None
                    or nm_profile.get_uuid() != self._nm_profile.get_uuid()
                )
            ):
                ProfileDelete(
                    self._ctx, self._iface.name, self._iface.type, nm_profile
                ).run()


class ProfileAdd:
    def __init__(
        self, ctx, iface_name, iface_type, nm_simple_conn, save_to_disk
    ):
        self._ctx = ctx
        self._iface_name = iface_name
        self._iface_type = iface_type
        self._nm_simple_conn = nm_simple_conn
        self._save_to_disk = save_to_disk

    def run(self):
        nm_add_conn2_flags = NM.SettingsAddConnection2Flags
        flags = nm_add_conn2_flags.BLOCK_AUTOCONNECT
        if self._save_to_disk:
            flags |= nm_add_conn2_flags.TO_DISK
        else:
            flags |= nm_add_conn2_flags.IN_MEMORY

        action = (
            f"Add profile: {self._nm_simple_conn.get_uuid()}, "
            f"iface:{self._iface_name}, type:{self._iface_type}"
        )
        self._ctx.register_async(action, fast=True)

        user_data = action
        args = None
        ignore_out_result = False  # Don't fall back to old AddConnection()
        self._ctx.client.add_connection2(
            self._nm_simple_conn.to_dbus(NM.ConnectionSerializationFlags.ALL),
            flags,
            args,
            ignore_out_result,
            self._ctx.cancellable,
            self._add_profile_callback,
            user_data,
        )

    def _add_profile_callback(self, nm_client, result, user_data):
        action = user_data
        if self._ctx.is_cancelled():
            return
        try:
            nm_profile = nm_client.add_connection2_finish(result)[0]
        except Exception as e:
            self._ctx.fail(
                NmstateLibnmError(f"{action} failed with error: {e}")
            )
            return

        if nm_profile is None:
            self._ctx.fail(
                NmstateLibnmError(
                    f"{action} failed with error: 'None returned from "
                    "NM.Client.add_connection2_finish()'"
                )
            )
        else:
            self._ctx.finish_async(action)


class ProfileUpdate:
    def __init__(
        self,
        ctx,
        iface_name,
        iface_type,
        nm_simple_conn,
        nm_profile,
        save_to_disk,
    ):
        self._ctx = ctx
        self._iface_name = iface_name
        self._iface_type = iface_type
        self._nm_simple_conn = nm_simple_conn
        self._nm_profile = nm_profile
        self._save_to_disk = save_to_disk

    def run(self):
        flags = NM.SettingsUpdate2Flags.BLOCK_AUTOCONNECT
        if self._save_to_disk:
            flags |= NM.SettingsUpdate2Flags.TO_DISK
        else:
            flags |= NM.SettingsUpdate2Flags.IN_MEMORY
        action = (
            f"Update profile uuid:{self._nm_profile.get_uuid()} "
            f"iface:{self._iface_name} type:{self._iface_type}"
        )
        user_data = action
        args = None

        self._ctx.register_async(action, fast=True)
        self._nm_profile.update2(
            self._nm_simple_conn.to_dbus(NM.ConnectionSerializationFlags.ALL),
            flags,
            args,
            self._ctx.cancellable,
            self._update_profile_callback,
            user_data,
        )

    def _update_profile_callback(self, nm_profile, result, user_data):
        action = user_data
        if self._ctx.is_cancelled():
            return
        try:
            ret = nm_profile.update2_finish(result)
        except Exception as e:
            self._ctx.fail(
                NmstateLibnmError(f"{action} failed with error={e}")
            )
            return

        if ret is None:
            self._ctx.fail(
                NmstateLibnmError(
                    f"{action} failed with error='None returned from "
                    "update2_finish()'"
                )
            )
        else:
            self._ctx.finish_async(action)


class ProfileDelete:
    def __init__(self, ctx, iface_name, iface_type, nm_profile):
        self._ctx = ctx
        self._iface_name = iface_name
        self._iface_type = iface_type
        self._nm_profile = nm_profile

    def run(self):
        action = (
            f"Delete profile: uuid:{self._nm_profile.get_uuid()} "
            f"id:{self._nm_profile.get_id()} "
            f"iface:{self._iface_name} type:{self._iface_type}"
        )
        user_data = action
        self._ctx.register_async(action, fast=True)
        self._nm_profile.delete_async(
            self._ctx.cancellable,
            self._delete_profile_callback,
            user_data,
        )

    def _delete_profile_callback(self, nm_profile, result, user_data):
        action = user_data
        if self._ctx.is_cancelled():
            return
        try:
            success = nm_profile.delete_finish(result)
        except Exception as e:
            self._ctx.fail(NmstateLibnmError(f"{action} failed: error={e}"))
            return

        if success:
            self._ctx.finish_async(action)
        else:
            self._ctx.fail(
                NmstateLibnmError(
                    f"{action} failed: error='None returned from "
                    "delete_finish'"
                )
            )


def _is_memory_only(nm_profile):
    if nm_profile:
        profile_flags = nm_profile.get_flags()
        return (
            NM.SettingsConnectionFlags.UNSAVED & profile_flags
            or NM.SettingsConnectionFlags.VOLATILE & profile_flags
        )
    return False


_NM_TYPE_2_NMSTATE = {
    NM.SETTING_WIRED_SETTING_NAME: InterfaceType.ETHERNET,
    NM.SETTING_BOND_SETTING_NAME: InterfaceType.BOND,
    NM.SETTING_DUMMY_SETTING_NAME: InterfaceType.DUMMY,
    NM.SETTING_TEAM_SETTING_NAME: InterfaceType.TEAM,
    NM.SETTING_VLAN_SETTING_NAME: InterfaceType.VLAN,
    NM.SETTING_VXLAN_SETTING_NAME: InterfaceType.VXLAN,
    NM.SETTING_BRIDGE_SETTING_NAME: InterfaceType.LINUX_BRIDGE,
    NM.SETTING_VRF_SETTING_NAME: InterfaceType.VRF,
    NM.SETTING_INFINIBAND_SETTING_NAME: InterfaceType.INFINIBAND,
    NM.SETTING_MACVLAN_SETTING_NAME: InterfaceType.MAC_VLAN,
    NM.SETTING_OVS_BRIDGE_SETTING_NAME: InterfaceType.OVS_BRIDGE,
    NM.SETTING_OVS_PORT_SETTING_NAME: InterfaceType.OVS_PORT,
    NM.SETTING_OVS_INTERFACE_SETTING_NAME: InterfaceType.OVS_INTERFACE,
}


def _get_iface_type(nm_profile):
    nm_iface_type = nm_profile.get_connection_type()
    iface_type = _NM_TYPE_2_NMSTATE.get(nm_iface_type, InterfaceType.UNKNOWN)
    if iface_type == InterfaceType.MAC_VLAN and is_macvtap(nm_profile):
        iface_type = InterfaceType.MAC_VTAP
    return iface_type


def _get_controller_info(nm_profile):
    nm_conn_setting = nm_profile.get_setting_connection()
    if nm_conn_setting and nm_conn_setting.props.master:
        return (
            nm_conn_setting.props.master,
            nm_conn_setting.props.slave_type,
        )
    return (None, None)


def get_basic_iface_info(nm_profile):
    info = {
        Interface.NAME: nm_profile.get_interface_name(),
        Interface.TYPE: _get_iface_type(nm_profile),
        Interface.STATE: InterfaceState.UP,
    }
    (
        info[BaseIface.CONTROLLER_METADATA],
        info[BaseIface.CONTROLLER_TYPE_METADATA],
    ) = _get_controller_info(nm_profile)
    return info



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from copy import deepcopy

from libnmstate.schema import Ethernet
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState
from libnmstate.ifaces.ethernet import EthernetIface

MAC_ADDRESS1 = "12:34:56:78:90:AB"
FOO_IFACE_NAME = "foo"


class TestEthernetIface:
    def test_merge_speed_duplex(self):
        iface_info = {
            Interface.NAME: FOO_IFACE_NAME,
            Interface.STATE: InterfaceState.UP,
            Interface.TYPE: InterfaceType.ETHERNET,
            Ethernet.CONFIG_SUBTREE: {
                Ethernet.AUTO_NEGOTIATION: False,
                Ethernet.SPEED: 1000,
                Ethernet.DUPLEX: Ethernet.FULL_DUPLEX,
            },
            Interface.MAC: MAC_ADDRESS1,
        }
        des_iface_info = {
            Interface.NAME: FOO_IFACE_NAME,
            Interface.TYPE: InterfaceType.ETHERNET,
        }

        cur_iface = EthernetIface(iface_info)
        des_iface = EthernetIface(des_iface_info)

        des_iface.merge(cur_iface)

        assert des_iface.to_dict() == {
            Interface.NAME: FOO_IFACE_NAME,
            Interface.STATE: InterfaceState.UP,
            Interface.TYPE: InterfaceType.ETHERNET,
            Ethernet.CONFIG_SUBTREE: {
                Ethernet.AUTO_NEGOTIATION: None,
                Ethernet.SPEED: None,
                Ethernet.DUPLEX: None,
            },
            Interface.MAC: MAC_ADDRESS1,
        }

    def test_mix_cases_of_vf_mac_address(self):
        mac_addr = "FF:EE:dd:CC:BB:aa"
        expected_mac_addr = mac_addr.upper()
        iface_info = {
            Interface.NAME: FOO_IFACE_NAME,
            Interface.STATE: InterfaceState.UP,
            Interface.TYPE: InterfaceType.ETHERNET,
            Ethernet.CONFIG_SUBTREE: {
                Ethernet.SRIOV_SUBTREE: {
                    Ethernet.SRIOV.TOTAL_VFS: 2,
                    Ethernet.SRIOV.VFS_SUBTREE: [
                        {
                            Ethernet.SRIOV.VFS.ID: 0,
                            Ethernet.SRIOV.VFS.SPOOF_CHECK: True,
                            Ethernet.SRIOV.VFS.MAC_ADDRESS: mac_addr,
                            Ethernet.SRIOV.VFS.TRUST: False,
                        }
                    ],
                },
            },
        }
        expected_iface_info = deepcopy(iface_info)
        expected_iface_info[Ethernet.CONFIG_SUBTREE][Ethernet.SRIOV_SUBTREE][
            Ethernet.SRIOV.VFS_SUBTREE
        ][0][Ethernet.SRIOV.VFS.MAC_ADDRESS] = expected_mac_addr

        iface = EthernetIface(iface_info)
        assert iface.state_for_verify() == expected_iface_info
